import pytest
import json
import time
from tmo import TmoClient


@pytest.fixture
def pat_tmo_client(mocker):
    mock_client = TmoClient.__new__(TmoClient)
    mock_client.ssl_verify = False
    mock_client.vmo_url = "https://awsmodelops.review.innovationlabs.teradata.com/api/accounts/mock_account/modelops/core"
    mock_client.auth_mode = "pat"
    mock_client._TmoClient__user = "pat_user"
    mock_client._TmoClient__pat = "pat_test_value"
    mock_client.session = mocker.MagicMock()
    mock_client.session.headers = {}
    mock_client.logger = mocker.Mock()

    return mock_client


def test_pat_authentication_sets_token(mocker, pat_tmo_client):
    mocker.patch.object(
        pat_tmo_client,
        "_TmoClient__create_self_signed_jwt",
        return_value="mock_auth_token",
    )
    mocker.patch("os.path.exists", return_value=False)

    pat_tmo_client._TmoClient__create_auth_session_pat()  # noqa

    assert pat_tmo_client.session.headers["Authorization"] == "Bearer mock_auth_token"


def test_token_is_reused_when_not_expired(mocker, pat_tmo_client):
    valid_token = {"token": "cached_token", "exp": time.time() + 180}
    m = mocker.mock_open(read_data=json.dumps(valid_token))
    mocker.patch("builtins.open", m)
    mocker.patch("os.path.exists", return_value=True)

    pat_tmo_client._TmoClient__create_auth_session_pat()  # noqa

    assert pat_tmo_client.session.headers["Authorization"] == "Bearer cached_token"


def test_token_is_refreshed_when_expired(mocker, pat_tmo_client):
    expired_token = {"token": "expired_token", "exp": time.time() - 10}
    m = mocker.mock_open(read_data=json.dumps(expired_token))
    mocker.patch("builtins.open", m)
    mocker.patch("os.path.exists", return_value=True)
    mocker.patch.object(pat_tmo_client, "_TmoClient__remove_cached_token")
    mocker.patch.object(
        pat_tmo_client,
        "_TmoClient__create_self_signed_jwt",
        return_value="new_mock_token",
    )

    pat_tmo_client._TmoClient__create_auth_session_pat()  # noqa

    assert pat_tmo_client.session.headers["Authorization"] == "Bearer new_mock_token"


def test_missing_pat_raises_exception(mocker, pat_tmo_client):
    pat_tmo_client._TmoClient__pat = None  # Simulate missing PAT

    with pytest.raises(Exception) as exc_info:
        pat_tmo_client._TmoClient__create_auth_session_pat()  # noqa

    assert "Missing CLI configuration" in str(exc_info.value)



from __future__ import absolute_import

from typing import Any

from tmo.api.iterator_base_api import IteratorBaseApi


class DatasetConnectionApi(IteratorBaseApi):
    path = "/api/datasetConnections"
    type = "DATASET_CONNECTION"

    def find_by_archived(
        self,
        archived: bool = False,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        raise NotImplementedError("Archiving not supported for DatasetConnections")

    def _get_header_params(self):
        return self._get_standard_header_params(
            accept_types=[
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]
        )

    def save(
        self,
        name: str,
        description: str,
        val_db: str,
        byom_db: str,
        password: str,
        personal: bool = True,
        creds_encrypted: bool = False,
        context: Any = None,
        log_mech: str = "TDNEGO",
        host: str = None,
        database: str = None,
        username: str = None,
    ):
        """
        register a dataset connection

        Parameters:
            name (str): dataset name
            description (str): dataset description
            val_db (str): val database
            byom_db (str): byom database
            password (str): password for Teradata Vantage connection
            personal (bool): flag to indicate if the dataset connection is personal
            creds_encrypted (bool): flag to indicate if credentials are encrypted
            context (any): teradataml context object containing connection information
            log_mech (str): log mechanism for Teradata Vantage connection. Default is "TDNEGO". Options are TDNEGO, TD2, LDAP, KRBS.
            host (str): Teradata Vantage host
            database (str): Teradata Vantage database
            username (str): Teradata Vantage username

        Raises:
            ValueError: If context is not provided and host, database, and username are not specified.
            HTTPError: If the request fails

        Returns:
            (dict): dataset template
        """

        if context:
            username = context.url.username
            host = context.url.host
            database = context.url.database
        else:
            if not all([host, database, username]):
                raise ValueError(
                    "If context is not provided, 'host', 'database', and 'username'"
                    " must be specified."
                )

        dataset_connection = {
            "name": name,
            "description": description,
            "metadata": {
                "host": host,
                "log_mech": log_mech,
                "database": database,
                "valDb": val_db,
                "byomDb": byom_db,
            },
            "credentials": {
                "username": username,
                "password": password,
                "credsEncrypted": creds_encrypted,
            },
            "personal": personal,
        }

        return self.tmo_client.post_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params={},
            body=dataset_connection,
        )

    def validate(
        self, dataset_connection: dict[str, Any], encrypted_credentials: bool = False
    ):
        """
        validate a dataset connection

        Parameters:
           dataset_connection (dict): dataset connection to validate
           encrypted_credentials (bool): flag to indicate if credentials are encrypted

        Returns:
            dict for resources, str for errors
        Raise:
            raises HTTPError in case of error status code
        """

        dataset_connection["credentials"]["credsEncrypted"] = encrypted_credentials

        return self.tmo_client.post_request(
            path=f"{self.path}/validate",
            header_params=self._get_header_params(),
            query_params={},
            body=dataset_connection,
        )



import io
import sys
import threading
from unittest import mock

import pytest

import tmo.spinner as spinner


def test_sys_print_writes_to_stdout(monkeypatch):
    buf = io.StringIO()
    monkeypatch.setattr(sys, "stdout", buf)
    spinner.sys_print("hello")
    assert buf.getvalue() == "hello"


def test_spin_it_invokes_spinner_and_returns(monkeypatch):
    calls = []

    class DummySpinner:
        def __init__(self, msg, speed):
            calls.append(("init", msg, speed))

        def start(self):
            calls.append("start")

        def stop(self):
            calls.append("stop")

        def remove_line(self):
            calls.append("remove_line")

    monkeypatch.setattr(spinner, "Spinner", DummySpinner)

    def sample_fn(x, y=2):
        calls.append(("fn", x, y))
        return "ok"

    result = spinner.spin_it(sample_fn, "testing", 0.123, 1, y=3)
    assert result == "ok"
    assert calls == [
        ("init", "testing", 0.123),
        "start",
        ("fn", 1, 3),
        "stop",
        "remove_line",
    ]


def test_progressbase_back_step_and_remove_line_calls_sys_print(monkeypatch):
    # Create instance without running __init__ to avoid threads
    pb = spinner.ProgressBase.__new__(spinner.ProgressBase)

    # Replace sys_print with a mock
    sp_mock = mock.Mock()
    monkeypatch.setattr(spinner, "sys_print", sp_mock)

    # When inplace is truthy, the sequences should be printed
    pb.inplace = 1
    pb.back_step()
    pb.remove_line()
    sp_mock.assert_any_call(spinner.CODE["CURSOR_NEXT_LINE"])
    sp_mock.assert_any_call(spinner.CODE["RM_LINE"])

    # Reset mock and ensure no calls when inplace is falsy
    sp_mock.reset_mock()
    pb.inplace = 0
    pb.back_step()
    pb.remove_line()
    sp_mock.assert_not_called()


def test_progressbase_start_calls_thread_start(monkeypatch):
    pb = spinner.ProgressBase.__new__(spinner.ProgressBase)
    mock_start = mock.Mock()
    monkeypatch.setattr(threading.Thread, "start", mock_start)
    pb.start()
    assert pb.stopFlag == 0
    mock_start.assert_called_once()


def test_progressbase_call_invokes_start(monkeypatch):
    pb = spinner.ProgressBase.__new__(spinner.ProgressBase)
    start_mock = mock.Mock()
    pb.start = start_mock
    pb()
    start_mock.assert_called_once()


def test_progressbase_stop_notifies_cv_and_acquires_rlock(monkeypatch):
    pb = spinner.ProgressBase.__new__(spinner.ProgressBase)
    pb.rlock = mock.Mock()
    pb.cv = mock.Mock()
    sp_mock = mock.Mock()
    monkeypatch.setattr(spinner, "sys_print", sp_mock)
    pb.stop()
    assert pb.stopFlag == 1
    sp_mock.assert_called_with(spinner.CODE["SHOW_CURSOR"])
    pb.cv.notify.assert_called_once()
    pb.rlock.acquire.assert_called_once()


def test_spinner_run_releases_rlock_and_prints(monkeypatch):
    s = spinner.Spinner("m", speed=0.001)
    s.rlock = mock.Mock()
    s.cv = mock.Mock()
    s.cv.acquire = mock.Mock()

    def wait_side_effect(timeout):
        if not hasattr(wait_side_effect, "calls"):
            wait_side_effect.calls = 1
            return None
        wait_side_effect.calls += 1
        s.stopFlag = 1
        return None

    s.cv.wait = mock.Mock(side_effect=wait_side_effect)
    sp_mock = mock.Mock()
    monkeypatch.setattr(spinner, "sys_print", sp_mock)
    s.inplace = 1
    s.stopFlag = 0
    s.run()
    sp_mock.assert_any_call(spinner.CODE["HIDE_CURSOR"])
    sp_mock.assert_any_call(spinner.CODE["CURSOR_NEXT_LINE"])
    s.rlock.release.assert_called_once()


def test_spinner_run_propagates_exception_and_does_not_release_rlock_on_error(
    monkeypatch,
):
    s = spinner.Spinner("m", speed=0.001)
    s.rlock = mock.Mock()
    s.cv = mock.Mock()
    s.cv.acquire = mock.Mock()
    s.cv.wait = mock.Mock(side_effect=RuntimeError("boom"))
    sp_mock = mock.Mock()
    monkeypatch.setattr(spinner, "sys_print", sp_mock)
    with pytest.raises(RuntimeError):
        s.run()
    assert not s.rlock.release.called


def test_spin_it_does_not_suppress_exceptions_from_function(monkeypatch):
    calls = []

    class DummySpinner:
        def __init__(self, msg, speed):
            calls.append(("init", msg, speed))

        def start(self):
            calls.append("start")

        def stop(self):
            calls.append("stop")

        def remove_line(self):
            calls.append("remove_line")

    monkeypatch.setattr(spinner, "Spinner", DummySpinner)

    def failing_fn():
        raise ValueError("fail")

    with pytest.raises(ValueError):
        spinner.spin_it(failing_fn, "msg", 0.01)
    assert ("start" in calls) and ("stop" not in calls)



import logging
from typing import Optional

import pandas
from teradataml import DataFrame

from tmo.types.dataset_metadata import Metadata

logger = logging.getLogger(__name__)


class BaseDatasetMixin:
    """Mixin that provides common functionality for Dataset and DatasetTemplate."""

    metadata: Metadata

    @property
    def entity(self) -> str:
        return self._entity

    @entity.setter
    def entity(self, value: str):
        if value is not None and not isinstance(value, str):
            raise ValueError("Entity must be a string.")
        self._entity = value

    @property
    def target(self) -> list[str]:
        return self._target

    @target.setter
    def target(self, value: list[str]):
        if value is not None and not isinstance(value, list):
            raise ValueError("Target must be a list of strings.")
        self._target = value

    @property
    def features_query(self) -> str:
        return self.metadata.features.sql

    @property
    def entity_targets_query(self) -> str:
        return self.metadata.entity_and_targets.sql

    @property
    def predictions_query(self) -> str:
        return self.metadata.predictions.entity_sql

    @property
    def features(self) -> list[str]:
        return [col.name for col in self.metadata.features.variables]

    def list_features(self) -> Optional[DataFrame | pandas.DataFrame]:
        """
        Returns the features in the dataset template as a DataFrame.

        Returns:
            Optional[teradataml.DataFrame | pandas.DataFrame]: A DataFrame containing feature information
        """
        from tmo.util.utils import to_dataframe
        from tmo.stats.stats_util import infer_columns_type

        if not hasattr(self, "metadata") or not hasattr(self.metadata, "features"):
            logger.error("Metadata or features not defined in the dataset template.")
            return None

        try:
            features = self.metadata.features.variables
            feature_columns = self.features
            features_query = self.metadata.features.sql

            categorical_features, _ = infer_columns_type(
                query=features_query,
                feature_columns=feature_columns,
            )

            feature_data = []
            for column in features:
                feature_type = (
                    "categorical"
                    if column.name in categorical_features
                    else "continuous"
                )
                feature_data.append({
                    "name": column.name,
                    "dataType": column.data_type,
                    "type": column.type,
                    "featureType": feature_type,
                })

            return to_dataframe(feature_data)
        except Exception as e:
            logger.error(f"Could not convert features to DataFrame: {str(e)}")
            return None

    def to_dataframe(self) -> Optional[DataFrame | pandas.DataFrame]:
        """
        Converts the dataset template to a DataFrame.

        Returns:
            Optional[teradataml.DataFrame | pandas.DataFrame]: A DataFrame representation of the dataset template.
        """
        from tmo.util.utils import to_dataframe

        try:
            return to_dataframe([self.get_df_template()])
        except Exception as e:
            logger.error(f"Could not convert dataset template to DataFrame: {str(e)}")
            return None

    @staticmethod
    def _isclass(obj, cls):
        return obj.__class__ == cls().__class__



import json
import os
import time

import aia
import oauthlib.oauth2.rfc6749.errors
import pandas as pd
import pytest
import requests

from tmo.api_client import TmoClient, ConfigurationError


def make_client(mocker):
    client = TmoClient.__new__(TmoClient)
    client.logger = mocker.Mock()
    client.ssl_verify = False
    client.vmo_url = None
    client.auth_mode = None
    client._TmoClient__client_id = None
    client._TmoClient__client_secret = None
    client._TmoClient__token_url = None
    client._TmoClient__device_auth_url = None
    client._TmoClient__bearer_token = None
    client._TmoClient__pat = None
    client._TmoClient__user = None
    client.project_id = None
    return client


def test_strip_url():
    assert (
        TmoClient._TmoClient__strip_url("https://example.com/")  # noqa
        == "https://example.com"
    )
    assert (
        TmoClient._TmoClient__strip_url("https://example.com")  # noqa
        == "https://example.com"
    )


def test_validate_and_extract_body_401_calls_remove_and_raises(mocker):
    client = make_client(mocker)
    # patch remove cached token to observe call
    mock_remove = mocker.patch.object(TmoClient, "_TmoClient__remove_cached_token")

    # response that simulates 401 and HTTPError on raise_for_status
    class Resp:
        status_code = 401
        text = "bad token text"

        def raise_for_status(self):
            raise requests.exceptions.HTTPError("original")

        def json(self):
            return {}

    resp = Resp()
    with pytest.raises(requests.exceptions.HTTPError) as exc:
        client._TmoClient__validate_and_extract_body(resp)  # noqa
    assert "Error message: bad token text" in str(exc.value)
    mock_remove.assert_called_once()


def test_create_self_signed_jwt_reads_key_and_saves(mocker, tmp_path):
    client = make_client(mocker)
    client.vmo_url = "https://org123.some.host"
    client._TmoClient__pat = "PATVALUE"
    client._TmoClient__user = "user1"

    pk_path = tmp_path / "pat.pem"
    pk_path.write_text("PRIVATE_KEY_CONTENT")
    # override default path so open reads our temp file
    client.DEFAULT_PAT_PRIVATE_KEY_PATH = str(pk_path)

    # patch jwt.encode and save method
    mock_encode = mocker.patch("tmo.api_client.jwt.encode", return_value="SIGNED_JWT")
    mock_save = mocker.patch.object(TmoClient, "_TmoClient__save_signed_jwt")

    signed = client._TmoClient__create_self_signed_jwt()  # noqa
    assert signed == "SIGNED_JWT"
    mock_encode.assert_called_once()
    # __create_self_signed_jwt should call __save_signed_jwt with signed jwt and exp
    mock_save.assert_called_once()
    args = mock_save.call_args[0]
    assert args[0] == "SIGNED_JWT"
    assert isinstance(args[1], int)


def test_remove_cached_token_deletes_file(tmp_path):
    # set class path to a temp file and ensure removal
    token_path = tmp_path / ".token"
    token_path.write_text("dummy")
    TmoClient.DEFAULT_TOKEN_CACHE_FILE_PATH = str(token_path)
    # call static method
    TmoClient._TmoClient__remove_cached_token()  # noqa
    assert not token_path.exists()


def test_validate_stored_token_no_file_generates_token_and_updates_session(mocker):
    client = make_client(mocker)
    # ensure no file exists
    mock_exists = mocker.patch("os.path.exists", return_value=False)
    # patch method that creates signed jwt
    mock_create_jwt = mocker.patch.object(
        TmoClient, "_TmoClient__create_self_signed_jwt", return_value="NEW_JWT"
    )
    # avoid TLS chase side effects by providing a dummy session class later
    if hasattr(client, "session"):
        del client.session

    client._TmoClient__validate_stored_token()  # noqa
    # after call, session and header should be present
    assert hasattr(client, "session")
    assert client.session.headers["Authorization"] == "Bearer NEW_JWT"
    mock_create_jwt.assert_called_once()


def test_validate_stored_token_expired_calls_remove_and_refresh(mocker, tmp_path):
    client = make_client(mocker)
    # create a token file with expired timestamp
    token_path = tmp_path / ".token"
    token_data = {"token": "OLD", "exp": int(time.time()) - 10}
    token_path.write_text(json.dumps(token_data))
    # patch DEFAULT_TOKEN_CACHE_FILE_PATH to point to this file
    TmoClient.DEFAULT_TOKEN_CACHE_FILE_PATH = str(token_path)

    mock_remove = mocker.patch.object(TmoClient, "_TmoClient__remove_cached_token")
    mock_create_jwt = mocker.patch.object(
        TmoClient, "_TmoClient__create_self_signed_jwt", return_value="REFRESHED"
    )
    # call validate
    client._TmoClient__validate_stored_token()  # noqa
    mock_remove.assert_called_once()
    assert client.session.headers["Authorization"] == "Bearer REFRESHED"
    mock_create_jwt.assert_called_once()


def test_validate_stored_token_valid_file_uses_existing_token(mocker, tmp_path):
    client = make_client(mocker)
    token_path = tmp_path / ".token"
    token_data = {"token": "GOOD", "exp": int(time.time()) + 3600}
    token_path.write_text(json.dumps(token_data))
    TmoClient.DEFAULT_TOKEN_CACHE_FILE_PATH = str(token_path)

    # no creation should happen
    mock_create_jwt = mocker.patch.object(
        TmoClient, "_TmoClient__create_self_signed_jwt"
    )
    client._TmoClient__validate_stored_token()  # noqa
    assert client.session.headers["Authorization"] == "Bearer GOOD"
    mock_create_jwt.assert_not_called()


def test_chase_tls_cert_chain_raises_when_no_vmo_url(mocker):
    client = make_client(mocker)
    # ensure vmo_url not set or falsy
    if hasattr(client, "vmo_url"):
        del client.vmo_url
    with pytest.raises(ConfigurationError):
        client._TmoClient__chase_tls_cert_chain()  # noqa


def test_set_session_tls_disables_warnings_when_ssl_false(mocker, monkeypatch):
    client = make_client(mocker)
    client.session = requests.Session()
    client.ssl_verify = False
    # make env var to something other than 'true' to trigger potential warning path
    monkeypatch.setenv("CALLED_FROM_TEST", "true")
    # patch disable_warnings
    disable_warnings = mocker.patch("requests.packages.urllib3.disable_warnings")
    # call method
    client._TmoClient__set_session_tls()  # noqa
    assert client.session.verify is False
    disable_warnings.assert_called_once()


def test_set_project_id_warns_if_project_not_found(mocker):
    client = make_client(mocker)
    client.logger = mocker.Mock()

    class FakeProjects:
        def find_by_id(self, project_id):
            return None

    mocker.patch.object(client, "projects", return_value=FakeProjects())
    client.set_project_id("nonexistent")
    # logger.warning should be called because project not found
    client.logger.warning.assert_called()


def test_create_auth_session_pat_raises_when_missing_credentials(mocker):
    # create instance and ensure pat/user are None
    client = TmoClient.__new__(TmoClient)
    client.logger = mocker.Mock()
    client._TmoClient__pat = None
    client._TmoClient__user = None
    with pytest.raises(ConfigurationError):
        client._TmoClient__create_auth_session_pat()  # noqa


def test_parse_kwargs_sets_attributes():
    client = TmoClient.__new__(TmoClient)
    client.vmo_url = None
    client.ssl_verify = True
    client.auth_mode = None
    client._TmoClient__client_id = None
    client._TmoClient__client_secret = None
    client._TmoClient__token_url = None
    client._TmoClient__device_auth_url = None
    client._TmoClient__bearer_token = None
    client._TmoClient__pat = None
    client._TmoClient__user = None
    kwargs = {
        "vmo_url": "https://test.com",
        "ssl_verify": False,
        "auth_mode": "pat",
        "pat": "PAT",
        "user": "user",
    }
    client._TmoClient__parse_kwargs(**kwargs)  # noqa
    assert client.vmo_url == "https://test.com"
    assert client.ssl_verify is False
    assert client.auth_mode == "pat"
    assert client._TmoClient__pat == "PAT"
    assert client._TmoClient__user == "user"


def test_parse_env_variables_sets_attributes(monkeypatch):
    client = TmoClient.__new__(TmoClient)
    monkeypatch.setenv("VMO_URL", "https://env.com")
    monkeypatch.setenv("VMO_SSL_VERIFY", "false")
    monkeypatch.setenv("VMO_API_AUTH_MODE", "bearer")
    monkeypatch.setenv("VMO_API_AUTH_BEARER_TOKEN", "TOKEN")
    client.vmo_url = None
    client.ssl_verify = True
    client.auth_mode = None
    client._TmoClient__bearer_token = None
    client._TmoClient__parse_env_variables()  # noqa
    assert client.vmo_url == "https://env.com"
    assert client.ssl_verify is False
    assert client.auth_mode == "bearer"
    assert client._TmoClient__bearer_token == "TOKEN"


def test_check_legacy_mode_oauth_cc():
    client = TmoClient.__new__(TmoClient)
    client.auth_mode = "oauth-cc"
    client._TmoClient__check_legacy_mode()  # noqa
    assert client.auth_mode == "client_credentials"


def test_check_legacy_mode_oauth():
    client = TmoClient.__new__(TmoClient)
    client.auth_mode = "oauth"
    client._TmoClient__check_legacy_mode()  # noqa
    assert client.auth_mode == "device_code"


def test_validate_url_raises():
    client = TmoClient.__new__(TmoClient)
    client.vmo_url = None
    with pytest.raises(ValueError):
        client._TmoClient__validate_url()  # noqa


def test_select_header_accept_none():
    assert TmoClient.select_header_accept([]) is None


def test_select_header_accept_list():
    result = TmoClient.select_header_accept(["application/json", "text/plain"])
    assert result == "application/json, text/plain"


def test_get_current_project():
    client = TmoClient.__new__(TmoClient)
    client.project_id = "abc123"
    assert client.get_current_project() == "abc123"


def test_init_raises_for_none(monkeypatch):
    monkeypatch.setattr(TmoClient, "_TmoClient__rename_tmo_config", lambda self: None)
    monkeypatch.setattr(
        TmoClient, "_TmoClient__parse_tmo_config", lambda self, **kwargs: None
    )
    with pytest.raises(ConfigurationError):
        TmoClient(auth_mode=None)


def test_init_raises_for_invalid(monkeypatch):
    monkeypatch.setattr(TmoClient, "_TmoClient__rename_tmo_config", lambda self: None)

    def _fake_parse(self, **kwargs):
        if "auth_mode" in kwargs:
            setattr(self, "auth_mode", kwargs.get("auth_mode"))

    monkeypatch.setattr(TmoClient, "_TmoClient__parse_tmo_config", _fake_parse)
    with pytest.raises(ValueError):
        TmoClient(auth_mode="invalid_mode")


def test_init_with_invalid_auth_mode():
    with pytest.raises(ValueError):
        TmoClient(auth_mode="invalid_mode")


def test_get_request_404_returns_none(mocker):
    client = make_client(mocker)
    client.vmo_url = "https://example.com"
    client.session = mocker.Mock()
    resp = mocker.Mock()
    resp.status_code = 404
    client.session.get.return_value = resp
    assert client.get_request("/path", {}, {}) is None


def test_get_request_retries_on_connection_error_then_success(mocker):
    client = make_client(mocker)
    client.vmo_url = "https://example.com"
    client.session = mocker.Mock()
    resp = mocker.Mock()
    resp.status_code = 200
    resp.json.return_value = {"ok": True}
    client.session.get.side_effect = [
        requests.exceptions.ConnectionError(),
        requests.exceptions.ConnectionError(),
        resp,
    ]
    result = client.get_request("/path", {}, {})
    assert result == {"ok": True}


def test_get_request_max_retries_logs_error_returns_none(mocker):
    client = make_client(mocker)
    client.vmo_url = "https://example.com"
    client.session = mocker.Mock()
    client.MAX_RETRIES = 2
    client.session.get.side_effect = requests.exceptions.ConnectionError()
    client.logger = mocker.Mock()
    ret = client.get_request("/path", {}, {})
    assert ret is None
    client.logger.error.assert_called()


def test_validate_and_extract_body_raises_original_http_error(mocker):
    client = make_client(mocker)

    class Resp:
        status_code = 500
        text = ""

        def raise_for_status(self):
            raise requests.exceptions.HTTPError("original")

        def json(self):
            return {}

    with pytest.raises(requests.exceptions.HTTPError) as exc:
        client._TmoClient__validate_and_extract_body(Resp())  # type: ignore
    assert "original" in str(exc.value)


def test_validate_and_extract_body_json_error_returns_text(mocker):
    client = make_client(mocker)

    class Resp:
        status_code = 200
        text = "plain text"

        def raise_for_status(self):
            return None

        def json(self):
            raise ValueError("no json")

    res = client._TmoClient__validate_and_extract_body(Resp())  # type: ignore
    assert res == "plain text"


def test_chase_tls_cert_chain_uses_aia_and_sets_verify(tmp_path, mocker):
    client = make_client(mocker)
    client.vmo_url = "https://org.example"
    client.session = requests.Session()
    mocker.patch("requests.get", side_effect=requests.exceptions.SSLError("ssl"))

    class FakeAIASession:
        def cadata_from_url(self, url):
            return "CERTDATA\n"

    mocker.patch("aia.AIASession", FakeAIASession)
    client._TmoClient__chase_tls_cert_chain()  # noqa
    assert isinstance(client.session.verify, str)
    assert os.path.exists(client.session.verify)
    os.remove(client.session.verify)


def test_chase_tls_cert_chain_invalid_ca_raises(tmp_path, mocker):
    client = make_client(mocker)
    client.vmo_url = "https://org.example"
    client.session = requests.Session()
    mocker.patch("requests.get", side_effect=requests.exceptions.SSLError("ssl"))

    class FakeAIASession:
        def cadata_from_url(self, url):
            raise aia.InvalidCAError("invalid ca")

    mocker.patch("aia.AIASession", FakeAIASession)
    with pytest.raises(requests.exceptions.SSLError):
        client._TmoClient__chase_tls_cert_chain()  # type: ignore


def test_set_session_tls_logs_warning_when_ssl_disabled(mocker):
    client = make_client(mocker)
    client.session = requests.Session()
    client.ssl_verify = False
    mocker.patch.dict(os.environ, {"CALLED_FROM_TEST": "false"}, clear=False)
    disable_warnings = mocker.patch("requests.packages.urllib3.disable_warnings")
    client._TmoClient__set_session_tls()  # type: ignore
    assert client.session.verify is False
    disable_warnings.assert_called_once()
    client.logger.warning.assert_called()


def test_create_oauth_session_client_credentials_missing_raises(mocker):
    client = make_client(mocker)
    client._TmoClient__client_id = None
    client._TmoClient__client_secret = None
    client._TmoClient__token_url = None
    with pytest.raises(ValueError):
        client._TmoClient__create_oauth_session_client_credentials()  # type: ignore


def test_create_oauth_session_bearer_sets_header(mocker):
    client = make_client(mocker)
    client._TmoClient__bearer_token = "Bearer TOKENX"
    client._TmoClient__create_oauth_session_bearer()  # type: ignore
    assert client.session.headers["Authorization"] == "Bearer TOKENX"


def test_get_device_code_non_200_raises(mocker):
    client = make_client(mocker)
    client._TmoClient__client_id = "cid"
    client._TmoClient__client_secret = None
    client._TmoClient__device_auth_url = "https://device"
    client._TmoClient__token_url = "https://token"
    client.session = mocker.Mock()
    resp = mocker.Mock()
    resp.status_code = 400
    resp.json.return_value = {}
    client.session.post.return_value = resp
    with pytest.raises(ValueError):
        client._TmoClient__get_device_code()  # type: ignore


def test_handle_device_code_retries_on_invalid_grant(mocker):
    client = make_client(mocker)
    err = oauthlib.oauth2.rfc6749.errors.InvalidGrantError(
        description="Token is not active"
    )
    called = {"count": 0}

    def fake_create(self):
        called["count"] += 1
        if called["count"] == 1:
            raise err
        return None

    mocker.patch.object(
        TmoClient, "_TmoClient__create_oauth_session_device_code", fake_create
    )
    mocker.patch.object(TmoClient, "_TmoClient__remove_cached_token")
    client._TmoClient__handle_device_code()  # type: ignore
    assert called["count"] == 2


def test_handle_device_code_retries_on_invalid_token_error(mocker):
    client = make_client(mocker)
    err = oauthlib.oauth2.rfc6749.errors.InvalidTokenError(description="Invalid token")
    called = {"count": 0}

    def fake_create(self):
        called["count"] += 1
        if called["count"] == 1:
            raise err
        return None

    mocker.patch.object(
        TmoClient, "_TmoClient__create_oauth_session_device_code", fake_create
    )
    mocker.patch.object(TmoClient, "_TmoClient__remove_cached_token")
    client._TmoClient__handle_device_code()  # type: ignore
    assert called["count"] == 2


def test_save_oauth_token_writes_and_sets_bearer(tmp_path, mocker):
    client = make_client(mocker)
    token_path = tmp_path / ".token"
    TmoClient.DEFAULT_TOKEN_CACHE_FILE_PATH = str(token_path)
    token = {"access_token": "abc123", "expires_in": 2}
    client._TmoClient__save_oauth_token(token)  # type: ignore
    with open(TmoClient.DEFAULT_TOKEN_CACHE_FILE_PATH, "r") as f:
        data = json.load(f)
    assert "expires_at" in data
    assert client._TmoClient__bearer_token == "Bearer abc123"  # type: ignore


def test_remove_cached_token_deletes_if_exists(tmp_path):
    path = tmp_path / ".token"
    path.write_text("x")
    TmoClient.DEFAULT_TOKEN_CACHE_FILE_PATH = str(path)
    TmoClient._TmoClient__remove_cached_token()  # type: ignore
    assert not path.exists()


# Additional tests merged from tests/api_client.py


def test_post_put_delete_request_calls_session_and_returns_page(mocker):
    client = make_client(mocker)
    client.vmo_url = "https://example.com/"
    client.auth_mode = "pat"
    client.session = mocker.Mock()
    mock_validate = mocker.patch.object(TmoClient, "_TmoClient__validate_stored_token")
    mock_validate_and_extract = mocker.patch.object(
        TmoClient, "_TmoClient__validate_and_extract_body", return_value={"ok": True}
    )

    body = {"a": 1}
    client.post_request("/p", {}, {}, body)  # type: ignore
    client.session.post.assert_called_once()
    called_args = client.session.post.call_args[1]
    assert called_args["url"] == "https://example.com/p"
    assert json.loads(called_args["data"]) == body

    client.session.post.reset_mock()
    client.put_request("/p2", {"h": "v"}, {"q": "1"}, body)  # type: ignore
    client.session.put.assert_called_once()
    put_args = client.session.put.call_args[1]
    assert put_args["url"] == "https://example.com/p2"
    assert json.loads(put_args["data"]) == body

    client.session.put.reset_mock()
    client.delete_request("/p3", {}, {}, body)  # type: ignore
    client.session.delete.assert_called_once()
    del_args = client.session.delete.call_args[1]
    assert del_args["url"] == "https://example.com/p3"
    assert json.loads(del_args["data"]) == body


def test_get_request_handles_unexpected_exception_and_logs(mocker):
    client = make_client(mocker)
    client.vmo_url = "https://example.com"
    client.session = mocker.Mock()
    client.session.get.side_effect = ValueError("boom")
    mock_log = mocker.patch("logging.error")
    res = client.get_request("/x", {}, {})
    assert res is None
    mock_log.assert_called()


def test_create_self_signed_jwt_raises_when_key_missing(mocker, tmp_path):
    client = make_client(mocker)
    client.vmo_url = "https://org123.something"
    client._TmoClient__pat = "p"
    client._TmoClient__user = "u"
    non_existing = tmp_path / "nope.pem"
    client.DEFAULT_PAT_PRIVATE_KEY_PATH = str(non_existing)
    with pytest.raises(FileNotFoundError):
        client._TmoClient__create_self_signed_jwt()  # type: ignore


def test_describe_current_project_returns_dataframe_when_project_found(mocker):
    client = make_client(mocker)
    client.project_id = "proj1"

    class FakeProjects:
        def find_by_id(self, project_id, expand=None):
            return {"id": project_id, "name": "P", "_links": {}, "userAttributes": {}}

    mocker.patch.object(client, "projects", return_value=FakeProjects())
    df = client.describe_current_project()
    assert isinstance(df, pd.DataFrame)
    assert "attribute" in df.columns and "value" in df.columns


def test_describe_current_project_returns_none_when_no_project_or_not_found(mocker):
    client = make_client(mocker)
    client.project_id = None
    assert client.describe_current_project() is None
    client.project_id = "p"

    class FakeProjects2:
        def find_by_id(self, project_id, expand=None):
            return None

    mocker.patch.object(client, "projects", return_value=FakeProjects2())
    assert client.describe_current_project() is None


def test_get_default_connection_id_variants(mocker):
    client = make_client(mocker)

    class UA:
        def get_default_connection(self):
            return {"value": {"defaultDatasetConnectionId": "cid123"}}

    mocker.patch.object(client, "user_attributes", return_value=UA())
    assert client.get_default_connection_id() == "cid123"

    class UA2:
        def get_default_connection(self):
            return None

    mocker.patch.object(client, "user_attributes", return_value=UA2())
    assert client.get_default_connection_id() is None

    class UA3:
        def get_default_connection(self):
            raise RuntimeError("boom")

    mocker.patch.object(client, "user_attributes", return_value=UA3())
    assert client.get_default_connection_id() is None


def test_api_wrapper_methods_return_clients(mocker):
    client = make_client(mocker)
    for method in (
        "projects",
        "datasets",
        "dataset_templates",
        "dataset_connections",
        "deployments",
        "feature_engineering",
        "jobs",
        "job_events",
        "messages",
        "models",
        "trained_models",
        "trained_model_artefacts",
        "trained_model_events",
        "user_attributes",
    ):
        func = getattr(client, method)
        obj = func()
        assert hasattr(obj, "tmo_client")
        assert obj.tmo_client is client


def test_get_device_code_successful_flow(mocker, monkeypatch):
    client = make_client(mocker)
    client._TmoClient__client_id = "cid"
    client._TmoClient__client_secret = None
    client._TmoClient__device_auth_url = "https://device"
    client._TmoClient__token_url = "https://token"
    client.session = mocker.Mock()

    device_resp = mocker.Mock()
    device_resp.status_code = 200
    device_resp.json.return_value = {
        "device_code": "DC",
        "user_code": "UC",
        "verification_uri_complete": "https://verify",
        "interval": 0,
    }

    token_pending = mocker.Mock()
    token_pending.status_code = 400
    token_pending.json.return_value = {"error": "authorization_pending"}

    token_ok = mocker.Mock()
    token_ok.status_code = 200
    token_ok.json.return_value = {"access_token": "ATOKEN", "expires_in": 3600}

    client.session.post.side_effect = [device_resp, token_pending, token_ok]

    monkeypatch.setattr("tmo.api_client.spin_it", lambda func, msg, speed: func())

    res = client._TmoClient__get_device_code()  # type: ignore
    assert isinstance(res, dict)
    assert res.get("access_token") == "ATOKEN"


def test_create_oauth_session_device_code_refreshes_with_refresh_token(
    mocker, tmp_path, monkeypatch
):
    client = make_client(mocker)
    client._TmoClient__client_id = "cid"
    client._TmoClient__client_secret = "secret"
    client._TmoClient__token_url = "https://token"
    client._TmoClient__device_auth_url = "https://device"
    client.ssl_verify = False

    token_path = tmp_path / ".token"
    expired = int(time.time()) - 3600
    token_file = {
        "expires_at": expired,
        "refresh_token": "RTOKEN",
        "access_token": "OLD",
    }
    token_path.write_text(json.dumps(token_file))
    client.DEFAULT_TOKEN_CACHE_FILE_PATH = str(token_path)

    class FakeOAuth2Session:
        def __init__(self, client_id=None):
            pass

        def refresh_token(
            self, token_url=None, refresh_token=None, auth=None, verify=None
        ):
            return {
                "access_token": "NEWAT",
                "expires_in": 3600,
                "refresh_token": "NEWRT",
            }

    monkeypatch.setattr("tmo.api_client.OAuth2Session", FakeOAuth2Session)
    monkeypatch.setattr("tmo.api_client.HTTPBasicAuth", lambda a, b: None)

    client._TmoClient__create_oauth_session_device_code()  # type: ignore

    assert client._TmoClient__bearer_token == "Bearer NEWAT"  # type: ignore


def test_get_device_code_authorization_declined_raises(mocker, monkeypatch):
    client = make_client(mocker)
    client._TmoClient__client_id = "cid"
    client._TmoClient__client_secret = None
    client._TmoClient__device_auth_url = "https://device"
    client._TmoClient__token_url = "https://token"
    client.session = mocker.Mock()

    device_resp = mocker.Mock()
    device_resp.status_code = 200
    device_resp.json.return_value = {
        "device_code": "DC",
        "user_code": "UC",
        "verification_uri_complete": "https://verify",
        "interval": 0,
    }

    token_declined = mocker.Mock()
    token_declined.status_code = 400
    token_declined.json.return_value = {
        "error": "authorization_declined",
        "error_description": "User declined authorization",
    }

    client.session.post.side_effect = [device_resp, token_declined]

    monkeypatch.setattr("tmo.api_client.spin_it", lambda func, msg, speed: func())

    from tmo.types.exceptions import AuthorizationError

    with pytest.raises(AuthorizationError) as exc:
        client._TmoClient__get_device_code()  # type: ignore
    assert "User declined authorization" in str(exc.value)


def test_get_device_code_bad_response_raises_authorization_error(mocker, monkeypatch):
    client = make_client(mocker)
    client._TmoClient__client_id = "cid"
    client._TmoClient__client_secret = None
    client._TmoClient__device_auth_url = "https://device"
    client._TmoClient__token_url = "https://token"
    client.session = mocker.Mock()

    device_resp = mocker.Mock()
    device_resp.status_code = 200
    device_resp.json.return_value = {
        "device_code": "DC",
        "user_code": "UC",
        "verification_uri_complete": "https://verify",
        "interval": 0,
    }

    token_bad = mocker.Mock()
    token_bad.status_code = 500
    token_bad.json.return_value = {}

    client.session.post.side_effect = [device_resp, token_bad]

    monkeypatch.setattr("tmo.api_client.spin_it", lambda func, msg, speed: func())

    from tmo.types.exceptions import AuthorizationError

    with pytest.raises(AuthorizationError) as exc:
        client._TmoClient__get_device_code()  # type: ignore
    assert "Bad response code 500" in str(exc.value)


def test_get_device_code_slow_down_calls_sleep_then_succeeds(mocker, monkeypatch):
    client = make_client(mocker)
    client._TmoClient__client_id = "cid"
    client._TmoClient__client_secret = None
    client._TmoClient__device_auth_url = "https://device"
    client._TmoClient__token_url = "https://token"
    client.session = mocker.Mock()

    device_resp = mocker.Mock()
    device_resp.status_code = 200
    device_resp.json.return_value = {
        "device_code": "DC",
        "user_code": "UC",
        "verification_uri_complete": "https://verify",
        "interval": 2,
    }

    token_slow = mocker.Mock()
    token_slow.status_code = 400
    token_slow.json.return_value = {"error": "slow_down"}

    token_ok = mocker.Mock()
    token_ok.status_code = 200
    token_ok.json.return_value = {"access_token": "ATOKEN", "expires_in": 3600}

    client.session.post.side_effect = [device_resp, token_slow, token_ok]

    sleep_mock = mocker.patch("tmo.api_client.time.sleep")
    monkeypatch.setattr("tmo.api_client.spin_it", lambda func, msg, speed: func())

    res = client._TmoClient__get_device_code()
    assert isinstance(res, dict)
    assert res.get("access_token") == "ATOKEN"
    sleep_mock.assert_called()



import logging
import os

from teradataml import create_context, get_context, configure
from typing_extensions import deprecated

logger = logging.getLogger(__name__)

SENSITIVE_ENV_VARS = [
    "VMO_CONN_PASSWORD",
    "VMO_CONN_USERNAME",
    "VMO_CONN_ENCRYPTION_TOKEN",
    "AOA_CONN_PASSWORD",
    "AOA_CONN_USERNAME",
    "AOA_CONN_ENCRYPTION_TOKEN",
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
]


@deprecated(
    "aoa_create_context is deprecated, please use tmo_create_context instead.",
    category=DeprecationWarning,
)
def aoa_create_context(database: str = None):
    """
    Creates a teradataml context if one does not already exist.
    Most users should not need to understand how we pass the environment variables etc. for dataset connections. This
    provides a way to achieve that and also allow them to work within a notebook for example where a context is already
    present.

    We create the connection based on the following environment variables which are configured automatically by ModelOps
    based on the dataset connection selected:

        AOA_CONN_HOST
        AOA_CONN_USERNAME
        AOA_CONN_PASSWORD
        AOA_CONN_LOG_MECH
        AOA_CONN_DATABASE
        AOA_VAL_INSTALL_DB
        AOA_BYOM_INSTALL_DB

    :param database: default database override
    :return: None
    """
    return tmo_create_context(database)


def tmo_create_context(database: str = None):
    """
    Creates a teradataml context if one does not already exist.
    Most users should not need to understand how we pass the environment variables etc. for dataset connections. This
    provides a way to achieve that and also allow them to work within a notebook for example where a context is already
    present.

    We create the connection based on the following environment variables which are configured automatically by ModelOps
    based on the dataset connection selected:

        VMO_CONN_HOST
        VMO_CONN_USERNAME
        VMO_CONN_PASSWORD
        VMO_CONN_LOG_MECH
        VMO_CONN_DATABASE
        VMO_VAL_INSTALL_DB
        VMO_BYOM_INSTALL_DB

    :param database: default database override
    :return: None
    """
    _rename_env_variables()

    try:
        if get_context() is None:
            if not database:
                database = os.getenv("VMO_CONN_DATABASE")

            host = os.environ["VMO_CONN_HOST"]
            logmech = os.getenv("VMO_CONN_LOG_MECH", "TDNEGO")
            username = os.environ["VMO_CONN_USERNAME"]
            password = os.environ["VMO_CONN_PASSWORD"]

            if database:
                logger.debug(
                    f"Configuring temp database for tables/views to {database}"
                )
                configure.temp_table_database = database
                configure.temp_view_database = database

            configure.val_install_location = os.environ.get("VMO_VAL_INSTALL_DB", "VAL")
            configure.byom_install_location = os.environ.get(
                "VMO_BYOM_INSTALL_DB", "MLDB"
            )

            logger.debug(
                f"Connecting to {host} on database {database} using logmech"
                f" {logmech} as {username}"
            )

            create_context(
                host=host,
                database=database,
                username=username,
                password=password,
                logmech=logmech,
            )

            from tmo import __version__

            execute_sql(f"""
            SET QUERY_BAND = 'appVersion={__version__};appName=VMO;appFunc=python;org=teradata-internal-telem;' FOR SESSION VOLATILE
            """)

        else:
            logger.info("teradataml context already exists. Skipping create_context.")
    finally:
        _sanitize_env_variables()


def execute_sql_in_context(context, statement, parameters=None):
    from teradataml.common.exceptions import TeradataMlException
    from teradataml.common.messages import Messages
    from teradataml.common.messagecodes import MessageCodes

    if context is not None:
        tdsql_con = context.raw_connection().driver_connection
        cursor = tdsql_con.cursor()
        return cursor.execute(statement, parameters)
    else:
        raise TeradataMlException(
            Messages.get_message(MessageCodes.INVALID_CONTEXT_CONNECTION),
            MessageCodes.INVALID_CONTEXT_CONNECTION,
        )


def execute_sql(statement, parameters=None):
    return execute_sql_in_context(get_context(), statement, parameters)


def _sanitize_env_variables():
    """
    Sanitize the environment variables to remove any sensitive information.
    """
    for key in os.environ:
        if key in SENSITIVE_ENV_VARS:
            del os.environ[key]


def _rename_env_variables():
    """
    Rename the environment variables from AOA to VMO.
    """
    for key in os.environ:
        if key.startswith("AOA_"):
            new_key = key.replace("AOA_", "VMO_")
            os.environ[new_key] = os.environ[key]



import logging
import uuid
from enum import Enum
from typing import Optional

from tmo.types.base_dataset import BaseDatasetMixin
from tmo.types.dataset_metadata import Metadata, CatalogType

logger = logging.getLogger(__name__)


class Scope(Enum):
    TRAIN = "train"
    EVALUATE = "evaluate"


class Dataset(BaseDatasetMixin):
    id: uuid.UUID
    dataset_template_id: uuid.UUID
    name: str
    description: str
    query: str
    scope: Scope
    metadata: Metadata
    catalog_type: CatalogType

    def __init__(
        self,
        dataset_template_id: Optional[uuid.UUID] = None,
        name: Optional[str] = "SDK Dataset",
        description: Optional[str] = "VANTAGE dataset",
        query: Optional[str] = None,
        scope: Optional[Scope] = None,
        metadata: Optional[Metadata] = None,
        catalog_type: CatalogType = CatalogType.VANTAGE,
    ):
        self.dataset_template_id = dataset_template_id
        self.name = name
        self.description = description
        self.query = query
        self.scope = scope
        self.metadata = metadata
        self.catalog_type = catalog_type

    @property
    def id(self) -> uuid.UUID:
        return self._id

    @id.setter
    def id(self, value: uuid.UUID):
        if value is not None and not isinstance(value, uuid.UUID):
            try:
                value = uuid.UUID(value)  # noqa
            except ValueError:
                raise ValueError("Project ID must be a valid UUID.")
        self._id = value  # noqa

    @property
    def dataset_template_id(self) -> uuid.UUID:
        return self._dataset_template_id

    @dataset_template_id.setter
    def dataset_template_id(self, value: uuid.UUID):
        if value is not None and not isinstance(value, uuid.UUID):
            try:
                value = uuid.UUID(value)  # noqa
            except ValueError:
                raise ValueError("Dataset template ID must be a valid UUID.")
        self._dataset_template_id = value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def description(self) -> str:
        return self._description

    @description.setter
    def description(self, value: str):
        self._description = value

    @property
    def query(self) -> str:
        return self._query

    @query.setter
    def query(self, value: str):
        if value is not None and not isinstance(value, str):
            raise ValueError("Query must be a string.")
        self._query = value

    @property
    def scope(self) -> Scope:
        return self._scope

    @scope.setter
    def scope(self, value: Scope):
        self._scope = value

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @metadata.setter
    def metadata(self, value: Metadata):
        self._metadata = value

    @property
    def catalog_type(self) -> CatalogType:
        return self._catalog_type

    @catalog_type.setter
    def catalog_type(self, value: CatalogType):
        if not isinstance(value, CatalogType):
            raise ValueError("Catalog type must be an instance of CatalogType.")
        self._catalog_type = value

    def get_df_template(self) -> dict:
        return {
            "datasetTemplateId": self.dataset_template_id,
            "name": self.name,
            "description": self.description,
            "catalogType": self.catalog_type,
            "scope": self.scope,
            "entity": self.entity,
            "target": self.target,
            "features": self.features,
            "featuresQuery": self.features_query,
            "entityTargetsQuery": self.entity_targets_query,
            "predictionsQuery": self.predictions_query,
        }



import json
import logging
import math
from decimal import Decimal
from typing import Optional, Any

import numpy as np
import pandas as pd
from teradataml import DataFrame  # noqa
from teradataml.analytics.valib import *  # noqa # NOSONAR(python:S2208)
from teradatasql import OperationalError

from tmo.stats import store

logger = logging.getLogger(__name__)

value_for_nan = 0
CHARSET_INVALID_PARAMETER = "'charset' is an invalid parameter"


class _NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating) or isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(_NpEncoder, self).default(obj)


def _compute_continuous_edges(
    variables: list[str], statistics: pd.DataFrame, dtypes: dict, bins=10
) -> list[float]:
    edges = []

    # Should return what is in the feature catalog, for now calculate linspace boundaries based on min/max
    ranges = statistics.drop(
        statistics.columns.difference(["xcol", "xmin", "xmax"]), axis=1
    )
    ranges = ranges.set_index("xcol")
    ranges.index = ranges.index.map(str.lower)
    ranges = ranges.to_dict(orient="index")

    for var in variables:
        x_min, x_max = ranges[var]["xmin"], ranges[var]["xmax"]

        # if integer type and range is less than the number of bins, only use 'range' bins
        if x_max - x_min < bins and dtypes[var].startswith("int"):
            edges.append(
                np.linspace(x_min, x_max, int(x_max) - int(x_min) + 1).tolist()
            )
        elif dtypes[var].startswith("decimal") or dtypes[var].startswith("float"):
            # (min / max must be rounded up / down accordingly so easier to just use the vals from stats)
            vals = np.linspace(x_min, x_max, bins + 1).tolist()

            # Rounding all edges to the best Teradata-compatible precision
            vals = list(map(_round_to_td_float, vals))

            dedup = []
            [dedup.append(x) for x in vals if x not in dedup]

            edges.append(dedup)
        else:  # Here we compute for regular int variables
            vals = np.linspace(x_min, x_max, bins + 1).tolist()
            for i in range(1, len(vals)):
                vals[i] = round(vals[i], 1)
            edges.append(vals)

    return edges


def _convert_all_edges_to_val_str(all_edges):
    # boundaries for multiple columns follows the following format..
    # ["{10, 0, 200000}", "{5, 0, 100}"]
    boundaries = []
    for edges in all_edges:
        edges_str = ",".join(str(edge) for edge in edges)
        boundaries.append("{{ {} }}".format(edges_str))

    return boundaries


def _fill_missing_bins(
    bin_edges: list[float], bin_values: list[float], reference_edges: list[float]
) -> list[float]:
    """
    Compare the `bin_edges` returned by VAL to `reference_edges` and fill in any missing bins with `0`.
    This is required as VAL doesn't return empty bins and so if we want to ensure we always have all the bins
    represented in the `bin_values` (which we do or else indices vs reference edges mean nothing), then we must do this.

    """
    new_bin_values = list(bin_values)

    epsilon = 1e-08
    for i, edge in enumerate(reference_edges):
        is_present = False
        for curr_edge in bin_edges:
            if abs(float(curr_edge) - float(edge)) < epsilon:
                is_present = True

        if not is_present:
            new_bin_values.insert(i, 0.0)

    return new_bin_values


def _extract_null_count(var_freq: pd.DataFrame) -> tuple[pd.DataFrame, float | None]:
    """
    Extract null count from frequency DataFrame and remove null rows.

    Parameters:
        var_freq: DataFrame containing frequency data for a variable

    Returns:
        tuple: (cleaned_var_freq, null_count) - DataFrame without nulls and null count
    """
    null_count = None

    if var_freq["xval"].isnull().values.any():
        null_rows = var_freq[var_freq["xval"].isnull()]
        null_count = null_rows.xcnt.tolist()[0]
        var_freq = var_freq[var_freq["xval"].notnull()]

    return var_freq, null_count


def _create_frequencies_dict(var_freq: pd.DataFrame) -> dict:
    """
    Create a dictionary mapping category values to their frequencies.

    Parameters:
        var_freq: DataFrame containing frequency data (without nulls)

    Returns:
        dict: mapping of category -> frequency count
    """
    if var_freq.empty:
        return {}

    return var_freq[["xval", "xcnt"]].set_index("xval").T.to_dict(orient="records")[0]


def _build_monitored_frequencies(
    frequencies_dict: dict,
    reference_categories: list,
    variable_name: str,
) -> tuple[dict, dict | None]:
    """
    Build monitored and unmonitored frequency dictionaries based on reference categories.

    Parameters:
        frequencies_dict: dict of all observed frequencies
        reference_categories: list of expected categories
        variable_name: name of the variable (for logging)

    Returns:
        tuple: (monitored_frequencies, unmonitored_frequencies)
    """
    # Build monitored frequencies (based on reference categories)
    monitored = {
        cat: frequencies_dict.get(cat, 0) if cat in reference_categories else 0
        for cat in reference_categories
    }

    # Identify new categories not in reference
    unmonitored = None
    for category in frequencies_dict.keys():
        if category not in reference_categories:
            logger.warning(
                f"Categorical feature {variable_name} has a new category {category} "
                "that's not listed in reference categories"
            )
            if unmonitored is None:
                unmonitored = {}
            unmonitored[category] = frequencies_dict.get(category)

    return monitored, unmonitored


def _process_categorical_var(
    frequencies: pd.DataFrame,
    group_label: str,
    variable_name: str,
    feature_importance: float,
    reference_categories: Optional[list],
) -> dict:
    """
    Process categorical variable statistics and build data structure.

    Creates a structure containing:
    - type: "categorical"
    - group: monitoring group name
    - importance: feature importance score
    - categories: reference categories list
    - statistics: nulls, frequency, and unmonitored_frequency

    Parameters:
        frequencies: DataFrame containing frequency data from VAL
        group_label: monitoring group name
        variable_name: name of the categorical variable
        feature_importance: importance score for this feature
        reference_categories: list of expected categories (None if unmonitored)

    Returns:
        dict: structured categorical variable statistics
    """
    # Initialize base structure
    data_struct: dict[str, str | float | dict[Any, Any] | list] = {
        "type": "categorical",
        "group": group_label,
        "statistics": {},
    }

    if feature_importance:
        data_struct["importance"] = feature_importance

    # Filter to this variable's frequencies
    var_freq = frequencies[frequencies.xcol == variable_name]

    # Extract and handle null values
    var_freq, null_count = _extract_null_count(var_freq)
    if null_count is not None:
        data_struct["statistics"]["nulls"] = null_count

    # Handle case where all values are null
    if var_freq.empty:
        logger.warning(
            f"Categorical feature {variable_name} has only NULL values, "
            "not computing a frequency"
        )
        return data_struct

    # Create frequency dictionary from DataFrame
    frequencies_dict = _create_frequencies_dict(var_freq)

    # Handle unmonitored features (no reference categories)
    if reference_categories is None:
        data_struct["statistics"]["unmonitored_frequency"] = frequencies_dict
        return data_struct

    # Record reference categories for monitored features
    data_struct["categories"] = reference_categories

    # Build monitored and unmonitored frequencies
    monitored, unmonitored = _build_monitored_frequencies(
        frequencies_dict, reference_categories, variable_name
    )

    data_struct["statistics"]["frequency"] = monitored

    if unmonitored is not None:
        data_struct["statistics"]["unmonitored_frequency"] = unmonitored

    return data_struct


def _process_continuous_var(
    histogram: pd.DataFrame,
    stats: pd.DataFrame,
    group_label: str,
    variable_name: str,
    feature_importance: float,
    reference_edges: Optional[list[float]],
) -> dict:
    """
    Adds the following struct
    {
        "type": "continuous",
        "group": "<group-name>",
        "importance": "<feature-importance>",
        "statistics": {
            "nulls": <missing-values>,
            "cnt": 614.0,
            "min": 21.0,
            "max": 81.0,
            ...
            "histogram": {
              "edges": [21.0,27.0,33.0,39.0,45.0,51.0,57.0,63.0,69.0,75.0,81.0],
              "values": [248.0,122.0,77.0,62.0,45.0,23.0,20.0,13.0,3.0,1.0]
            }
        }
    }
    """

    data_struct: dict[str, str | float | dict[Any, Any] | list] = {
        "type": "continuous",
        "group": group_label,
        "statistics": {},
    }

    if feature_importance:
        data_struct["importance"] = feature_importance

    def _strip_key_x(d: dict):
        return {k[1:]: v for k, v in d.items()}

    stats_values = (
        stats[stats.xcol == variable_name]
        .drop(["xdb", "xtbl", "xcol"], axis=1)
        .to_dict(orient="records")[0]
    )
    data_struct["statistics"].update(_strip_key_x(stats_values))

    if histogram.empty:
        # No histogram available - we stop here, not reporting histogram
        return data_struct

    var_hist = histogram[histogram.xcol == variable_name].sort_values(by=["xbin"])

    # if first row is nan then it is the null values in the dataset. remove from histogram
    if var_hist["xbin"].isnull().values.any():
        n = var_hist[var_hist["xbin"].isnull()]
        data_struct["statistics"]["nulls"] = n.xcnt.tolist()[0]

        var_hist = var_hist[var_hist["xbin"].notnull()]

    if var_hist.empty:
        # We should probably generate all-zero histogram here?
        logger.warning(
            f"Continuous feature {variable_name} has only NULL values, not computing a"
            " histogram"
        )
        return data_struct

    bin_edges = [var_hist.xbeg.tolist()[0]] + var_hist.xend.tolist()
    bin_values = var_hist.xcnt.tolist()

    # (issue #123) VAL docs originally stated that:
    # VAL histograms will add values lower than the first bin to the first bin, but values greater than the
    # largest bin are added to a new bin. Therefore, we did the same on both sides. However, it turns out this doc is
    # incorrect.

    is_right_outlier_bin = math.isnan(bin_edges[-1])
    is_left_outlier_bin = math.isnan(bin_edges[0])
    if is_right_outlier_bin:
        bin_edges = bin_edges[:-1]
    if is_left_outlier_bin:
        bin_edges = bin_edges[1:]

    # Add missing bin_values based on the bin_edges vs reference_edges.
    # VAL doesn't return empty bins
    if len(bin_edges) < len(reference_edges):
        bin_values = _fill_missing_bins(
            bin_edges=bin_edges, bin_values=bin_values, reference_edges=reference_edges
        )

    right_outliers = 0
    left_outliers = 0
    if is_right_outlier_bin:
        right_outliers = int(bin_values[-1])
        logger.warning(
            f"Continuous feature {variable_name} has values larger than the rightmost"
            f" bin, identified {right_outliers} values"
        )
        bin_values[-2] += bin_values[
            -1
        ]  # Probably should consider removing as some point, when outliers are properly reported and visible to the user
        bin_values = bin_values[:-1]
    if is_left_outlier_bin:
        left_outliers = int(bin_values[0])
        logger.warning(
            f"Continuous feature {variable_name} has values smaller than the leftmost"
            f" bin, identified {left_outliers} values"
        )
        bin_values[1] += bin_values[
            0
        ]  # Probably should consider removing as some point, when outliers are properly reported and visible to the user
        bin_values = bin_values[1:]

    data_struct["statistics"]["histogram"] = {
        "edges": reference_edges,
        "values": bin_values,
    }

    if is_right_outlier_bin:
        data_struct["statistics"]["histogram"]["right_outliers"] = right_outliers
    if is_left_outlier_bin:
        data_struct["statistics"]["histogram"]["left_outliers"] = left_outliers

    return data_struct


def _validate_dataframe_type(df, df_name: str):
    """
    Validate that the provided dataframe is a teradataml DataFrame.

    Parameters:
        df: dataframe to validate
        df_name (str): name of the dataframe for error message

    Raises:
        TypeError: if df is not a teradataml DataFrame
    """
    if not isinstance(df, DataFrame):
        raise TypeError(f"We only support teradataml DataFrame for {df_name}")


def _ensure_backward_compatibility(data_stats: dict) -> dict:
    """
    Ensure backward compatibility by renaming 'predictors' to 'targets' if present.

    Parameters:
        data_stats (dict): data statistics dictionary

    Returns:
        dict: updated data statistics
    """
    # Ensure backward compatible (when we had targets incorrectly named as predictors)
    if "predictors" in data_stats:
        data_stats["targets"] = data_stats.pop("predictors")
    return data_stats


def _extract_variable_lists(data_stats: dict) -> tuple[list[str], list[str], list[str]]:
    """
    Extract and categorize variables from data statistics.

    Parameters:
        data_stats (dict): data statistics dictionary containing features and targets

    Returns:
        tuple: (features, targets, categorical) - three lists of variable names
    """
    features = []
    targets = []
    categorical = []

    for var_type in ["features", "targets"]:
        if var_type not in data_stats:
            continue

        for name, value in data_stats[var_type].items():
            # For backward compatibility with data stats created before we lower-case
            name = name.lower()

            if var_type == "features":
                features.append(name)
            elif var_type == "targets":
                targets.append(name)

            if "type" in value and value["type"] == "categorical":
                categorical.append(name)

    return features, targets, categorical


def _validate_matching_row_counts(predicted_df: DataFrame, features_df: DataFrame):
    """
    Validate that prediction and feature dataframes have matching row counts.

    Parameters:
        predicted_df: predictions dataframe
        features_df: features dataframe

    Raises:
        ValueError: if row counts don't match
    """
    if predicted_df.shape[0] != features_df.shape[0]:
        raise ValueError(
            "The number of prediction rows do not match the number of features rows!"
        )


def _parse_scoring_stats(
    features_df: DataFrame,
    predicted_df: DataFrame,
    data_stats: dict,
    feature_importance: dict[str, float] = None,
    feature_metadata_fqtn: str = None,
    feature_metadata_group: str = None,
) -> dict:
    """
    Parse scoring statistics from features and predictions dataframes.

    Parameters:
        features_df: teradataml DataFrame containing features
        predicted_df: teradataml DataFrame containing predictions
        data_stats (dict): existing data statistics
        feature_importance (dict): feature importance scores
        feature_metadata_fqtn (str): fully qualified table name for feature metadata
        feature_metadata_group (str): monitoring group name

    Returns:
        dict: combined statistics for features and targets
    """
    feature_importance = feature_importance or {}

    # Validate input dataframes
    _validate_dataframe_type(features_df, "features")
    _validate_dataframe_type(predicted_df, "predictions")

    # Ensure backward compatibility
    data_stats = _ensure_backward_compatibility(data_stats)

    # Extract variable lists
    features, targets, categorical = _extract_variable_lists(data_stats)

    # Validate matching row counts
    _validate_matching_row_counts(predicted_df, features_df)

    # Capture statistics for features
    data_stats = _capture_stats(
        df=features_df,
        features=features,
        targets=[],
        categorical=categorical,
        feature_importance=feature_importance,
        feature_metadata_fqtn=feature_metadata_fqtn,
        feature_metadata_group=feature_metadata_group,
    )

    # Capture statistics for targets
    targets_struct = _capture_stats(
        df=predicted_df,
        features=[],
        targets=targets,
        categorical=categorical,
        feature_importance=feature_importance,
        feature_metadata_fqtn=feature_metadata_fqtn,
        feature_metadata_group=feature_metadata_group,
    )

    # Combine features and targets statistics
    data_stats["targets"] = targets_struct["targets"]

    return data_stats


def _normalize_variable_names(
    features: list,
    targets: list,
    categorical: list,
    feature_importance: dict,
) -> tuple[list, list, list, dict]:
    """
    Normalize all variable names to lowercase for case-insensitive matching.

    Parameters:
        features: list of feature names
        targets: list of target names
        categorical: list of categorical variable names
        feature_importance: dict of feature importance scores

    Returns:
        tuple: (normalized_features, normalized_targets, normalized_categorical, normalized_importance)
    """
    features = [f.lower() for f in features]
    targets = [t.lower() for t in targets]
    categorical = [c.lower() for c in categorical]
    feature_importance = {k.lower(): v for k, v in feature_importance.items()}

    return features, targets, categorical, feature_importance


def _validate_variables_in_dataframe(
    features: list, targets: list, df_columns: list, original_columns: list
):
    """
    Validate that features and targets exist in the DataFrame.

    Parameters:
        features: list of feature names (lowercase)
        targets: list of target names (lowercase)
        df_columns: list of DataFrame column names (lowercase)
        original_columns: original DataFrame columns for error message

    Raises:
        ValueError: if features or targets are not in DataFrame
    """
    if features and not set(features).issubset(df_columns):
        raise ValueError(
            f"features dataframe with columns ({original_columns}) does not contain"
            f" features: {features}"
        )

    if targets and not set(targets).issubset(df_columns):
        raise ValueError(
            f"targets dataframe with columns ({original_columns}) does not contain"
            f" targets: {targets}"
        )


def _categorize_variables(
    features: list, targets: list, categorical: list
) -> tuple[list, list]:
    """
    Categorize variables into continuous and categorical types.

    Parameters:
        features: list of feature names
        targets: list of target names
        categorical: list of categorical variable names

    Returns:
        tuple: (continuous_vars, categorical_vars)
    """
    continuous_vars = list((set(features) | set(targets)) - set(categorical))
    categorical_vars = list((set(features) | set(targets)) - set(continuous_vars))

    return continuous_vars, categorical_vars


def _process_continuous_variables(
    df: DataFrame,
    continuous_vars: list,
    feature_metadata_fqtn: str,
) -> tuple[pd.DataFrame | None, pd.DataFrame | None, list, list]:
    """
    Process continuous variables and retrieve statistics and histograms.

    Parameters:
        df: teradataml DataFrame
        continuous_vars: list of continuous variable names
        feature_metadata_fqtn: fully qualified table name for feature metadata

    Returns:
        tuple: (stats, histogram, reference_edges, available_continuous_vars)
    """
    if len(continuous_vars) == 0:
        return None, None, [], []

    stats = _safe_compute_valib_statistics(df, continuous_vars)
    stats = stats.result.to_pandas().fillna(value_for_nan).reset_index()
    stats["xcol"] = stats["xcol"].str.lower()

    stats_metadata = store.get_feature_stats(feature_metadata_fqtn, "continuous")

    reference_edges = []
    available_continuous_vars = []

    for v in continuous_vars:
        if v.lower() in stats_metadata.keys():
            reference_edges.append(stats_metadata[v.lower()]["edges"])
            available_continuous_vars.append(v)
        else:
            logger.warning(
                f"Feature {v} doesn't have statistics metadata defined, and will not be"
                " monitored.\nIn order to enable monitoring for this feature, make"
                " sure that statistics metadata is available in"
                f" {feature_metadata_fqtn}"
            )

    histogram = None
    if len(available_continuous_vars) > 0:
        hist_boundaries = _convert_all_edges_to_val_str(reference_edges)
        histogram = _histogram_wrapper(
            data=df, columns=available_continuous_vars, boundaries=hist_boundaries
        )
        histogram["xcol"] = histogram["xcol"].str.lower()

    return stats, histogram, reference_edges, available_continuous_vars


def _process_categorical_variables(
    df: DataFrame,
    categorical_vars: list,
    feature_metadata_fqtn: str,
) -> tuple[pd.DataFrame | None, list, list]:
    """
    Process categorical variables and retrieve frequencies.

    Parameters:
        df: teradataml DataFrame
        categorical_vars: list of categorical variable names
        feature_metadata_fqtn: fully qualified table name for feature metadata

    Returns:
        tuple: (frequencies, reference_categories, available_categorical_vars)
    """
    if len(categorical_vars) == 0:
        return None, [], []

    stats_metadata = store.get_feature_stats(feature_metadata_fqtn, "categorical")

    reference_categories = []
    available_categorical_vars = []

    for v in categorical_vars:
        if v.lower() in stats_metadata.keys():
            reference_categories.append(stats_metadata[v.lower()]["categories"])
            available_categorical_vars.append(v)
        else:
            logger.warning(
                f"Feature {v} doesn't have statistics metadata defined, and will not be"
                " monitored.\nIn order to enable monitoring for this feature, make"
                " sure that statistics metadata is available in"
                f" {feature_metadata_fqtn}"
            )

    # Compute frequencies for all categorical features
    frequencies = _safe_compute_valib_frequency(df, categorical_vars)
    frequencies = frequencies.result.to_pandas(all_rows=True).reset_index()
    frequencies["xcol"] = frequencies["xcol"].str.lower()

    return frequencies, reference_categories, available_categorical_vars


def _build_variable_metadata(
    variable_name: str,
    continuous_vars: list,
    available_continuous_vars: list,
    available_categorical_vars: list,
    stats: pd.DataFrame | None,
    histogram: pd.DataFrame | None,
    frequencies: pd.DataFrame | None,
    reference_edges: list,
    reference_categories: list,
    group_label: str,
    feature_importance: dict,
) -> dict:
    """
    Build metadata structure for a single variable.

    Parameters:
        variable_name: name of the variable
        continuous_vars: list of continuous variables
        available_continuous_vars: list of continuous vars with metadata
        available_categorical_vars: list of categorical vars with metadata
        stats: statistics DataFrame for continuous variables
        histogram: histogram DataFrame for continuous variables
        frequencies: frequencies DataFrame for categorical variables
        reference_edges: list of reference edges for continuous variables
        reference_categories: list of reference categories for categorical variables
        group_label: monitoring group name
        feature_importance: dict of feature importance scores

    Returns:
        dict: variable metadata structure
    """
    if variable_name in continuous_vars:
        if variable_name in available_continuous_vars:
            return _process_continuous_var(
                histogram=histogram,
                stats=stats,
                reference_edges=reference_edges[
                    available_continuous_vars.index(variable_name)
                ],
                group_label=group_label,
                variable_name=variable_name,
                feature_importance=feature_importance.get(variable_name, 0.0),
            )
        else:
            return _process_continuous_var(
                histogram=pd.DataFrame(),
                stats=stats,
                reference_edges=None,
                group_label=group_label,
                variable_name=variable_name,
                feature_importance=feature_importance.get(variable_name, 0.0),
            )
    else:  # categorical variable
        if variable_name in available_categorical_vars:
            return _process_categorical_var(
                frequencies=frequencies,
                group_label=group_label,
                variable_name=variable_name,
                feature_importance=feature_importance.get(variable_name, 0.0),
                reference_categories=reference_categories[
                    available_categorical_vars.index(variable_name)
                ],
            )
        else:
            return _process_categorical_var(
                frequencies=frequencies,
                group_label=group_label,
                variable_name=variable_name,
                feature_importance=feature_importance.get(variable_name, 0.0),
                reference_categories=None,
            )


def _capture_stats(
    df: DataFrame,
    features: list,
    targets: list,
    categorical: list,
    feature_importance: dict[str, float] = None,
    feature_metadata_fqtn: str = None,
    feature_metadata_group: str = "default",
) -> dict:
    """
    Capture statistics for features and targets from a DataFrame.

    Parameters:
        df: teradataml DataFrame containing data
        features: list of feature names
        targets: list of target names
        categorical: list of categorical variable names
        feature_importance: dict of feature importance scores
        feature_metadata_fqtn: fully qualified table name for feature metadata
        feature_metadata_group: monitoring group name

    Returns:
        dict: statistics structure containing features and targets metadata
    """
    feature_importance = feature_importance or {}

    if not isinstance(df, DataFrame):
        raise TypeError("We only support teradataml DataFrame")

    if not feature_metadata_fqtn:
        raise ValueError("feature_metadata_fqtn must be defined")

    # Normalize all variable names to lowercase
    original_columns = df.columns
    df_columns = [c.lower() for c in df.columns]
    features, targets, categorical, feature_importance = _normalize_variable_names(
        features, targets, categorical, feature_importance
    )

    # Validate that variables exist in DataFrame
    _validate_variables_in_dataframe(features, targets, df_columns, original_columns)

    # Categorize variables into continuous and categorical
    continuous_vars, categorical_vars = _categorize_variables(
        features, targets, categorical
    )

    # Process continuous variables
    stats, histogram, reference_edges, available_continuous_vars = (
        _process_continuous_variables(df, continuous_vars, feature_metadata_fqtn)
    )

    # Process categorical variables
    frequencies, reference_categories, available_categorical_vars = (
        _process_categorical_variables(df, categorical_vars, feature_metadata_fqtn)
    )

    # Build data structure
    total_rows = df.shape[0]
    data_struct = {"num_rows": total_rows, "features": {}, "targets": {}}

    # Add metadata for each feature
    for variable_name in features:
        data_struct["features"][variable_name] = _build_variable_metadata(
            variable_name=variable_name,
            continuous_vars=continuous_vars,
            available_continuous_vars=available_continuous_vars,
            available_categorical_vars=available_categorical_vars,
            stats=stats,
            histogram=histogram,
            frequencies=frequencies,
            reference_edges=reference_edges,
            reference_categories=reference_categories,
            group_label=feature_metadata_group,
            feature_importance=feature_importance,
        )

    # Add metadata for each target
    for variable_name in targets:
        data_struct["targets"][variable_name] = _build_variable_metadata(
            variable_name=variable_name,
            continuous_vars=continuous_vars,
            available_continuous_vars=available_continuous_vars,
            available_categorical_vars=available_categorical_vars,
            stats=stats,
            histogram=histogram,
            frequencies=frequencies,
            reference_edges=reference_edges,
            reference_categories=reference_categories,
            group_label=feature_metadata_group,
            feature_importance=feature_importance,
        )

    # Track missing metadata
    missing_continuous = [
        c for c in continuous_vars if c.lower() not in available_continuous_vars
    ]
    missing_categorical = [
        c for c in categorical_vars if c.lower() not in available_categorical_vars
    ]

    if len(missing_categorical) > 0 or len(missing_continuous) > 0:
        data_struct["missing_metadata"] = {}
        if len(missing_continuous) > 0:
            data_struct["missing_metadata"]["continuous"] = missing_continuous
        if len(missing_categorical) > 0:
            data_struct["missing_metadata"]["categorical"] = missing_categorical

    return data_struct


def _safe_compute_valib_frequency(features_df: DataFrame, features: list[str]):
    try:
        logger.debug(
            "Executing this VAL"
            f" {valib.Frequency(data=features_df, columns=features, charset='UTF8').show_query()}"
        )
        frequencies = valib.Frequency(
            data=features_df, columns=features, charset="UTF8"
        )
    except OperationalError as opError:
        if CHARSET_INVALID_PARAMETER in str(opError):
            logger.debug(
                "Executing this VAL"
                f" {valib.Frequency(data=features_df, columns=features).show_query()}"
            )
            frequencies = valib.Frequency(data=features_df, columns=features)
        else:
            raise

    return frequencies


def _safe_compute_valib_statistics(features_df: DataFrame, features: list[str]):
    try:
        logger.debug(
            "Executing this VAL"
            f" {valib.Statistics(data=features_df, columns=features, stats_options='all', charset='UTF8').show_query()}"
        )
        statistics = valib.Statistics(
            data=features_df, columns=features, stats_options="all", charset="UTF8"
        )
    except OperationalError as opError:
        if CHARSET_INVALID_PARAMETER in str(opError):
            logger.debug(
                "Executing this VAL"
                f" {valib.Statistics(data=features_df, columns=features, stats_options='all').show_query()}"
            )
            statistics = valib.Statistics(
                data=features_df, columns=features, stats_options="all"
            )
        else:
            raise

    return statistics


def _safe_compute_valib_histogram(
    features_df: DataFrame, features: list[str], boundaries: list[str | float | int]
):
    try:
        logger.debug(
            "Executing this VAL"
            f" {valib.Histogram(data=features_df,columns=features,boundaries=boundaries, charset='UTF8').show_query()}"
        )
        histogram = valib.Histogram(
            data=features_df, columns=features, boundaries=boundaries, charset="UTF8"
        )
    except OperationalError as opError:
        if CHARSET_INVALID_PARAMETER in str(opError):
            logger.debug(
                "Executing this VAL"
                f" {valib.Histogram(data=features_df,columns=features,boundaries=boundaries).show_query()}"
            )
            histogram = valib.Histogram(
                data=features_df, columns=features, boundaries=boundaries
            )
        else:
            raise

    return histogram


def _histogram_wrapper(data, columns, boundaries):
    working_columns = list(columns)
    working_boundaries = list(boundaries)
    result = []

    # Iterate over an updated list of columns while it's not empty
    while working_columns:
        tc = list(working_columns)
        tb = list(working_boundaries)
        # If the size of td_analyze parameter is bigger than the limit
        # then reduce the list of columns until it fits
        while _val_parameter_length_too_long(tc, tb):
            if len(tc) == 1:
                raise ValueError(
                    "VAL histogram parameters are too long even for single column"
                    f" {tc[0]}"
                )
            tc.pop()
            tb.pop()

        # compute a histogram on the longest possible list of columns
        histogram = _safe_compute_valib_histogram(data, tc, tb)
        result.append(histogram.result.to_pandas(all_rows=True).reset_index())

        # removing columns thats already computed
        working_columns = working_columns[len(tc) :]
        working_boundaries = working_boundaries[len(tb) :]

    return pd.concat(result)


def _val_parameter_length_too_long(columns, boundaries, limit=31000):
    db_tbl_string = "database=;tablename=;outputdatabase=;outputtablename=;"
    column_string = f"columns={','.join(columns)};boundaries={','.join(boundaries)};"
    # assuming maximum length for database and table names at 128
    final_length = len(db_tbl_string) + len(column_string) + 128 * 4
    if final_length > limit:
        logger.debug(
            "VAL histogram parameters are too long for these columns, trying to"
            f" compute sequentially: {','.join(columns)}"
        )
    return final_length > limit


def _round_to_td_float(x, max_digits=15):
    # This function is needed to cast/round decimal and float number to the maximum precision allowed by Teradata database.
    # Teradata floats allow maximum 15 digits for mantissa, so this strange logic below does exactly that.
    # Possibly at some point we may need to add a round to exponent as well.
    (sign, digits, exponent) = Decimal(x).as_tuple()
    e = len(digits) + exponent
    (_, m, _) = (
        Decimal(x)
        .scaleb(-e)
        .quantize(Decimal(f"1E-{max_digits}"))
        .normalize()
        .as_tuple()
    )
    return np.float64(Decimal((sign, m, e - len(m))))


def infer_columns_type(
    df=None, query=None, feature_columns=None, cardinality_threshold=10
):

    if query:
        df = DataFrame.from_query(query)
        df = df.to_pandas()

    categorical_features = []
    continuous_features = []

    for col in feature_columns:
        try:
            unique_values = df[col].dropna().unique()
            num_unique = len(unique_values)

            if df[col].dtype == "object" or df[col].dtype.name == "category":
                categorical_features.append(col)

            elif num_unique == 2 and all(val in [0, 1] for val in unique_values):
                categorical_features.append(col)

            elif (
                pd.api.types.is_numeric_dtype(df[col])
                and num_unique <= cardinality_threshold
            ):
                categorical_features.append(col)

            elif pd.api.types.is_numeric_dtype(df[col]):
                continuous_features.append(col)
            else:
                continuous_features.append(col)

        except Exception as e:
            print(f"Error processing column '{col}': {e}. Skipping.")
            continue

    return categorical_features, continuous_features



import inspect
import logging

from typing import Optional, Literal, Any, Union

logger = logging.getLogger(__name__)


class functional(object):  # noqa @ NOSONAR
    """
    A class decorator that converts class attributes to properties with getter and setter methods following the functional programming paradigm.
    It also adds class-level getter, setter, and deleter methods to manage multiple attributes at once.

    Example:
        @functional
        class Point:
            x = 0
            y = 0

        # Create instance
        p = Point(1, 2)
        print(p.x)  # 1
        print(p.y)  # 2

        # Use functional getters and setters
        p.set_x(10)
        print(p.get_x())  # 10

        # Get all attributes as a dict
        print(p.get("dict"))  # {'x': 10, 'y': 2}

        # Change several attributes at once
        p.set(x=5, y=7)
        print(p.get())  # [5, 7]

        # Delete attributes
        p.del(["x"])
        # print(p.x)  # AttributeError

    Usage:
        - Decorate a class with @functional to enable functional access to its attributes.
        - Use get_<attr>(), set_<attr>(), del_<attr>() methods to manipulate individual attributes.
        - Use get(), set(), del() methods to manipulate multiple attributes.
    """

    _cls = None
    _instance_obj = None
    _init_args = []
    _final_args = {}

    def __init__(self, *args, **kwargs) -> None:
        self._has_init = True if args[0].__init__ is not object.__init__ else False
        self._init_args = [
            x for x in inspect.getfullargspec(args[0].__init__).args if x != "self"
        ]
        if (
            len(args) == 0
            or len(kwargs) > 0
            or ((len(args) > 0) and not inspect.isclass(args[0]))
        ):
            raise TypeError(
                f"@{self.__class__.__name__} decorator can only be used to decorate"
                " classes and does not take any arguments."
            )

        self._cls = args[0]

    def __call__(self, *args, **kwargs):
        if len(args) > 0 and len(kwargs) > 0:
            raise TypeError(
                f"@{self.__class__.__name__} {self._cls.__name__} can only take either"
                " positional or keyword arguments, not both."
            )

        if (
            self._has_init is False
            and len(kwargs) == 0
            and len(args) > len(self._get_cls_attrs(self._cls))
        ):
            raise TypeError(
                f"@{self.__class__.__name__} {self._cls.__name__} takes at most"
                f" {len(self._get_cls_attrs(self._cls))} positional arguments but"
                f" {len(args)} were given."
            )

        self._args = args
        self._kwargs = kwargs
        self._final_args = dict(zip(self._init_args, self._args))

        if len(args) > 0 and len(self._init_args) > 0:
            self._instance_obj = self._cls(*args)
        elif len(kwargs) > 0 and self._has_init is True:
            self._instance_obj = self._cls(**kwargs)
        else:
            self._instance_obj = self._cls()

        return self._process_class()

    def _process_class(self) -> object:
        """Processes the class and adds properties and getter/setter methods for each attribute, then initializes the attributes values depending on the default values and the values passed in the builder."""
        attr_index = 0
        for attr, value in self._get_cls_attrs(self._cls).items():
            self._instance_obj = self._add_properties(self._instance_obj, attr, value)
            self._instance_obj = self._value_initializer(
                self._instance_obj, attr, value, attr_index
            )
            attr_index += 1

        return self._functional_cls_getter_setter(self._instance_obj)

    def _add_properties(self, cls=None, name=None, attr=None) -> object:
        """Adds a private property to the class for the specified attribute name and also adds getter, setter, and deleter methods."""
        setattr(cls, f"_{name}", attr)  # store the actual value in a private attribute
        prop = property(
            fget=lambda selfie: getattr(cls, f"_{name}"),
            fset=lambda selfie, value: setattr(cls, f"_{name}", value),
            fdel=lambda selfie: delattr(cls, f"_{name}"),
            doc=f"Property for {name}",
        )
        setattr(cls, name, prop)

        return self._functional_attr_getter_setter(cls, name)

    def _functional_attr_getter_setter(self, cls: object, name: str) -> object:
        """Adds getter, setter, and deleter methods to the class for the specified attribute name."""
        self._add_attr_getter(cls, name)
        self._add_attr_setter(cls, name)
        self._add_attr_deleter(cls, name)
        return cls

    def _functional_cls_getter_setter(self, cls: object) -> object:
        """Adds class-level getter, setter, and deleter methods to the class."""
        self._add_cls_getter(cls)
        self._add_cls_setter(cls)
        self._add_cls_deleter(cls)
        return cls

    def _value_initializer(
        self, cls: object, name: str, attr: Any, index: int = 0
    ) -> object:
        """Initializes the attribute value based on the provided positional or keyword arguments."""
        if len(self._final_args) > 0:
            if name in self._final_args.keys():
                setattr(cls, name, self._final_args[name])
            else:
                setattr(cls, name, attr)
        elif len(self._args) > 0:
            if index < len(self._args):
                setattr(cls, name, self._args[index])
            else:
                setattr(cls, name, attr)
        elif len(self._kwargs) > 0 and name in self._kwargs:
            setattr(cls, name, self._kwargs[name])
        else:
            setattr(cls, name, attr)
        return cls

    @staticmethod
    def _add_attr_getter(cls: object, name: str) -> object:
        """Adds a getter method to the class for the specified attribute name."""

        def wrapper():
            return getattr(cls, name)

        setattr(cls, f"get_{name}", wrapper)
        setattr(
            getattr(cls, f"get_{name}"),
            "__doc__",
            f"""Functional getter for {cls.__class__.__name__}.{name}""",
        )
        return wrapper

    @staticmethod
    def _add_attr_setter(cls: object, name: str) -> object:
        """Adds a setter method to the class for the specified attribute name."""

        def wrapper(value):
            setattr(cls, name, value)
            return cls

        setattr(cls, f"set_{name}", wrapper)
        setattr(
            getattr(cls, f"set_{name}"),
            "__doc__",
            f"""Functional setter for {cls.__class__.__name__}.{name}""",
        )
        return wrapper

    @staticmethod
    def _add_attr_deleter(cls: object, name: str) -> object:
        """Adds a deleter method to the class for the specified attribute name."""

        def wrapper():
            delattr(cls, name)
            return cls

        setattr(cls, f"del_{name}", wrapper)
        setattr(
            getattr(cls, f"del_{name}"),
            "__doc__",
            f"""Functional deleter for {cls.__class__.__name__}.{name}""",
        )
        setattr(cls, f"delete_{name}", wrapper)
        setattr(
            getattr(cls, f"delete_{name}"),
            "__doc__",
            f"""Functional deleter for {cls.__class__.__name__}.{name}""",
        )
        return wrapper

    @staticmethod
    def _add_cls_getter(cls: object) -> object:
        """Adds an attributes getter method to the class. Takes a list of attribute names to get and an output type (dict or list)."""

        def wrapper(
            attributes: Optional[
                Union[list[str], Literal["dict", "list", "all"]]
            ] = None,
            output: Literal["dict", "list"] = "list",
        ):
            if attributes in ["dict", "list"]:
                output = attributes
                attributes = None
            if attributes is None or len(attributes) == 0 or attributes == "all":
                attributes = [
                    attr for attr, val in functional._get_cls_attrs(cls).items()
                ]
            if output == "dict":
                return {attr: getattr(cls, attr) for attr in attributes}
            elif output == "list":
                return [getattr(cls, attr) for attr in attributes]
            else:
                raise ValueError("Output type must be either 'dict' or 'list'")

        setattr(cls, "get", wrapper)
        setattr(
            getattr(cls, "get"),
            "__doc__",
            f"""Functional getter for {cls.__class__.__name__}""",
        )
        return wrapper

    @staticmethod
    def _add_cls_setter(cls: object) -> object:
        """Adds an attributes setter method to the class. Takes either a list of positional arguments or keyword arguments to set the attributes."""

        def wrapper(*args, **kwargs):
            if len(args) > 0 and len(kwargs) > 0:
                raise TypeError(
                    "set() can only take either positional or keyword arguments, not"
                    " both."
                )
            if len(args) > 0:
                attributes = [
                    attr for attr, name in functional._get_cls_attrs(cls).items()
                ]
                if len(args) > len(attributes):
                    raise ValueError(
                        "Number of arguments exceeds number of attributes in the class."
                    )
                for i, value in enumerate(args):
                    setattr(cls, attributes[i], value)
            elif len(kwargs) > 0:
                functional._add_csl_setter_kwargs(cls, kwargs)
            return cls

        setattr(cls, "set", wrapper)
        setattr(
            getattr(cls, "set"),
            "__doc__",
            f"""Functional getter for {cls.__class__.__name__}""",
        )
        return wrapper

    @staticmethod
    def _add_csl_setter_kwargs(cls: object, kwargs) -> object:
        for key, value in kwargs.items():
            if key not in functional._get_cls_attrs(cls).keys():
                raise AttributeError(
                    f"Attribute {key} not found in class {cls.__class__.__name__}"
                )
            setattr(cls, key, value)

    @staticmethod
    def _add_cls_deleter(cls: object) -> object:
        """Adds an attributes deleter method to the class. Takes a list of attribute names to delete."""

        def wrapper(
            attributes: Optional[Union[list[str], Literal["all"]]] = None,
        ) -> object:
            if attributes is None or len(attributes) == 0:
                raise ValueError("Attributes list cannot be None or empty")
            cls_attrs = [attr for attr, val in functional._get_cls_attrs(cls).items()]
            if attributes == "all":
                attributes = cls_attrs
            for attr in attributes:
                if attr not in cls_attrs:
                    raise ValueError(
                        f"Attribute {attr} not found in class {cls.__class__.__name__}"
                    )
                delattr(cls, attr)
                delattr(cls, f"_{attr}")
            return cls

        setattr(cls, "del", wrapper)
        setattr(
            getattr(cls, "del"),
            "__doc__",
            f"""Functional deleter for {cls.__class__.__name__}""",
        )
        setattr(cls, "delete", wrapper)
        setattr(
            getattr(cls, "delete"),
            "__doc__",
            f"""Functional deleter for {cls.__class__.__name__}""",
        )
        return wrapper

    @staticmethod
    def _get_cls_attrs(cls: object) -> dict:
        """Returns a dictionary of class attributes excluding dunder and sunder methods, callable and private attributes or methods, and 'self'."""
        return {
            name: attr
            for name, attr in cls.__dict__.items()
            if not (
                (name.startswith("__") and name.endswith("__")) or name.startswith("_")
            )
            and not callable(attr)
            and attr != "self"
        }



import logging
import os
import shutil

ARTIFACTS_PATH = "./artifacts"
MODELS_PATH = "./models"

logger = logging.getLogger(__name__)


def cleanup_artifacts():
    """
    Clean up artifacts and models directories.

    Removes the artifacts directory if it exists and the models
    file/directory if it exists.
    """
    if os.path.exists(ARTIFACTS_PATH):
        # Handle symlinks separately as shutil.rmtree fails on them
        if os.path.islink(ARTIFACTS_PATH):
            os.unlink(ARTIFACTS_PATH)
        else:
            shutil.rmtree(ARTIFACTS_PATH)
    try:
        os.remove(MODELS_PATH)
    except OSError:
        logger.debug(f"Nothing to remove, folder '{MODELS_PATH}' does not exists.")



import inspect
import logging
import os
import runpy
import shutil
import subprocess
import types

from tmo.cli.base_task import BaseTask
from tmo.context.model_context import ModelContext, DatasetInfo

ARTIFACTS_PATH = "./artifacts"


class RunTask(BaseTask):

    def __init__(self, repo_manager):
        super().__init__(repo_manager)
        self.logger = logging.getLogger(__name__)

    def run_task_local(self, base_path: str, task_name: str = None, func: str = None):

        base_path = os.path.join(base_path, "")
        task_path = base_path + "feature_engineering_tasks/"

        if not os.path.exists(task_path):
            raise ValueError(
                f"feature engineering task directory {task_path} does not exist"
            )

        catalog = BaseTask.get_task_folders(task_path)  # noqa

        cli_task_kwargs = self._BaseTask__get_task_varargs()  # noqa

        if not task_name or task_name not in catalog:
            print("")
            print(
                "Task name not found. Please select one from the list below or"
                " press Ctrl+C to quit."
            )
            task_name = self.__input_select(
                "task",
                catalog,
                "Available feature engineering tasks:",
            )

        task_file_path = os.path.join(task_path, task_name, "task.py")

        task_artefacts_path = f".artefacts/{task_name}/"  # noqa
        task_artefacts_abs_path = os.path.abspath(task_artefacts_path)
        task_artefacts_output_path = os.path.join(task_artefacts_path, "output/")
        if os.path.exists(task_artefacts_path):
            self.logger.debug(
                f"Cleaning local task artefact path: {task_artefacts_path}"
            )
            shutil.rmtree(task_artefacts_path)

        os.makedirs(task_artefacts_output_path)

        try:
            if os.path.exists(ARTIFACTS_PATH):
                os.remove(ARTIFACTS_PATH)

            if os.name == "nt":
                subprocess.run(
                    ["cmd", "/c", "mklink", "/J", "artifacts", task_artefacts_abs_path],
                    check=True,
                    capture_output=True,
                    text=True,
                )
            else:
                os.symlink(
                    task_artefacts_path, ARTIFACTS_PATH, target_is_directory=True
                )

            context = ModelContext(
                artifact_output_path="artifacts/output",
                dataset_info=DatasetInfo.from_dict({}),
                hyperparams={},
                **cli_task_kwargs,
            )

            task_module = runpy.run_path(task_file_path)

            if not func or not (
                func in task_module
                and isinstance(task_module[func], types.FunctionType)
            ):
                function_names = [
                    name
                    for name, value in task_module.items()
                    if isinstance(value, types.FunctionType)
                    and inspect.getfile(value) == task_file_path
                ]
                print("")
                print(
                    "Function name not found. Please select one from the list below or"
                    " press Ctrl+C to quit."
                )
                func = self.__input_select(
                    "function", function_names, "Available functions:"
                )

            print("")

            self.logger.info("Loading and executing task code")
            task_module[func](context=context)

            self.logger.info(f"Artefacts can be found in: {task_artefacts_output_path}")
            self.__cleanup()
            return task_name, func
        except ModuleNotFoundError:
            self.__cleanup()
            self.logger.error(
                "Missing required python module, try running the following command"
                " first:"
            )
            self.logger.error(f"pip install -r {task_path}requirements.txt")
            raise
        except:
            self.__cleanup()
            self.logger.exception("Exception running feature engineering task code")
            raise

    @staticmethod
    def __cleanup():
        if os.path.exists(ARTIFACTS_PATH):
            os.remove(ARTIFACTS_PATH)

    @staticmethod
    def __print_underscored(message):  # TODO: should be moved to utils
        print(message)
        print("-" * len(message))

    def __input_select(
        self, name, values, label="", default=None
    ) -> str | None:  # TODO: should be moved to utils
        if len(values) == 0:
            return None

        if label != "":
            self.__print_underscored(label)

        for ix, item in enumerate(values):  # noqa
            default_text = " (default)" if default and default == item else ""
            print(f"[{ix}] {item}" + default_text)

        tmp_index = (
            input(f"Select {name} by index (or leave blank for the default one): ")
            if default
            else input(f"Select {name} by index: ")
        )

        if default and default in values and tmp_index == "":
            tmp_index = values.index(default)
        elif (tmp_index == "" and not default) or (
            not tmp_index.isnumeric() or int(tmp_index) >= len(values)
        ):
            print(
                "Wrong selection, please try again by selecting the index number on the"
                " first column."
            )
            print("You may cancel at anytime by pressing Ctrl+C.")
            print("")
            return self.__input_select(name, values, label, default)

        return values[int(tmp_index)]



from __future__ import absolute_import

import logging

from tmo.api.iterator_base_api import IteratorBaseApi

logger = logging.getLogger(__name__)


class ProjectApi(IteratorBaseApi):

    path = "/api/projects"
    type = "PROJECT"

    def _get_header_params(self) -> dict:
        # The header for project id is required for the archive/unarchive method from base_api
        return self._get_standard_header_params()

    def find_by_name_like(self, name: str, projection: str = None) -> list:
        """
        returns a list of projects matching the given name

        Parameters:
           name (str): project name to search by
           projection (str): projection type

        Returns:
            (list): list of projects
        """
        query_vars = ["name", "projection"]
        query_vals = [name, projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/search/findByName",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def save(self, project: dict[str, str | bool | dict]) -> dict:
        """
        create a project

        Parameters:
           project (dict): project to create

        Returns:
            (dict): project
        """
        header_vars = ["Accept"]
        header_vals = [self.json_type]
        header_params = self.generate_params(header_vars, header_vals)

        project.setdefault("archived", False)
        project.setdefault("isChecked", False)
        project.setdefault("gitCredentials", {"source": "none"})

        return self.tmo_client.post_request(
            path=self.path, header_params=header_params, query_params={}, body=project
        )

    def update(self, project: dict[str, str]) -> dict:
        """
        update a project

        Parameters:
           project (dict): project to update

        Returns:
            (dict): project
        """
        header_vars = ["Accept"]
        header_vals = [self.json_type]
        header_params = self.generate_params(header_vars, header_vals)

        return self.tmo_client.put_request(
            path=f"{self.path}/{self.tmo_client.project_id}",
            header_params=header_params,
            query_params={},
            body=project,
        )

    def is_archived(self, project_id: str) -> bool | None:
        """
        returns the archived status of a project
        Parameters:
           project_id (str): project id to check archived status
        Returns:
            (bool)
        """
        project = self.find_by_id(project_id)

        if not project or "archived" not in project:
            logger.warning(f"Project with id {project_id} not found")
            return None

        return project["archived"]



from __future__ import absolute_import

from tmo.api.iterator_base_api import IteratorBaseApi


class TrainedModelEventApi(IteratorBaseApi):

    path = "/api/trainedModelEvents"
    type = "TRAINED_MODEL_EVENT"

    def _get_header_params(self):
        return self._get_standard_header_params(
            accept_types=[
                self.json_type,
                "application/hal+json",
                "text/uri-list",
                "application/x-spring-data-compact+json",
            ]
        )

    def find_all(
        self,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns all trained model events

        Parameters:
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): all trained model events
        """
        query_vars = ["projection", "page", "sort", "size", "sort"]
        query_vals = [projection, page, sort, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=self.path,
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_id(self, event_id: str, projection: str = None):
        """
        returns a trained model event

        Parameters:
           event_id (str): job id(uuid) to find
           projection (str): projection type

        Returns:
            (dict): trained model event
        """
        query_vars = ["projection"]
        query_vals = [projection]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/{event_id}",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_status(
        self,
        status: str,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns trained model events by status

        Parameters:
           status (str): status of trained model event
            Available values : TRAINED, EVALUATED, APPROVED, REJECTED, CHAMPION, DEPLOYED, RETIRED
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): trained model events
        """
        query_vars = ["status", "projection", "page", "sort", "size", "sort"]
        query_vals = [status, projection, page, sort, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/search/findByStatus",
            header_params=self._get_header_params(),
            query_params=query_params,
        )

    def find_by_trained_model_id(
        self,
        trained_model_id: str,
        projection: str = None,
        page: int = None,
        size: int = None,
        sort: str = None,
    ):
        """
        returns trained model events by trained model id

        Parameters:
           trained_model_id (str): trained model id(uuid)
           projection (str): projection type
           page (int): page number
           size (int): number of records in a page
           sort (str): column name and sorting order
           e.g. name?asc: sort name in ascending order, name?desc: sort name in descending order

        Returns:
            (dict): trained model events
        """
        query_vars = ["trainedModelId", "projection", "page", "sort", "size", "sort"]
        query_vals = [trained_model_id, projection, page, sort, size, sort]
        query_params = self.generate_params(query_vars, query_vals)

        return self.tmo_client.get_request(
            path=f"{self.path}/search/findByTrainedModelId",
            header_params=self._get_header_params(),
            query_params=query_params,
        )



import json
import logging
import os
import shutil
from pathlib import Path

import yaml
from git import Repo

from tmo.types.entity import EntityType
from .evaluate_model import EvaluateModel
from .score_model import ScoreModel
from .train_model import TrainModel

MODEL_JSON = "model.json"
FEATURE_ENGINEERING_PATH = "feature_engineering_tasks"
MODEL_DEFINITIONS_PATH = "model_definitions"


class RepoManager(object):

    def __init__(self, base_path: str):
        self.base_path = base_path
        self.logger = logging.getLogger(__name__)

    def add_model(
        self,
        model_id: str,
        model_name: str,
        model_desc: str,
        template: str,
        base_path: str = None,
    ):

        base_path = self.base_path if base_path is None else base_path
        model_path = os.path.join(base_path, MODEL_DEFINITIONS_PATH)

        if not os.path.isdir(model_path):
            try:
                os.makedirs(model_path)
            except Exception as e:
                raise IOError(f"Could not create models directory {model_path}: {e}")

        with open(os.path.join(template, MODEL_JSON), "r") as f:
            model = json.load(f)
            model["source_model_id"] = model["id"]
            model["id"] = model_id
            model["name"] = model_name
            model["description"] = model_desc

            name_path = model_name.strip().lower().replace(" ", "_")

            target_path = os.path.join(model_path, name_path)
            if os.path.isdir(target_path):
                raise ValueError(
                    f"Model path {name_path} already exists, please try another model"
                    " name"
                )

            try:
                shutil.copytree(template, target_path)
            except Exception as e:
                raise IOError(f"Could not copy model template to {target_path}: {e}")

            with open(os.path.join(target_path, MODEL_JSON), "w") as t:
                json.dump(model, t, indent=4)

    def add_task(self, template: str, task_name: str, base_path: str = None):

        base_path = self.base_path if base_path is None else base_path
        task_path = os.path.join(base_path, FEATURE_ENGINEERING_PATH)

        if not os.path.exists(task_path):
            try:
                os.makedirs(task_path)
            except Exception as e:
                raise IOError(
                    "Could not create feature engineering tasks directory"
                    f" {task_path}: {e}"
                )

        name_path = task_name.strip().lower().replace(" ", "_")

        target_path = os.path.join(task_path, name_path)
        if os.path.isdir(target_path):
            raise ValueError(
                f"Task path {name_path} already exists, please try another task name"
            )

        try:
            shutil.copytree(template, target_path)
        except Exception as e:
            raise IOError(f"Could not copy model template to {target_path}: {e}")

    def get_templates(
        self, entity_type: EntityType = EntityType.MODEL, source_path: str = None
    ) -> dict:

        source_path = (
            self.base_path if source_path is None else os.path.join(source_path, "")
        )
        templates_path = os.path.join(
            source_path,
            (
                MODEL_DEFINITIONS_PATH
                if entity_type == EntityType.MODEL
                else FEATURE_ENGINEERING_PATH
            ),
        )
        templates = {}

        if not os.path.isdir(templates_path):
            return templates

        if entity_type == EntityType.MODEL:
            templates = RepoManager._get_models_from_repo(templates_path)
        elif entity_type == EntityType.FEATURE_ENGINEERING_TASK:
            templates = RepoManager._get_tasks_from_repo(templates_path)

        return templates

    def init_model_directory(self, path: str = None):
        logging.info("Creating model directory")
        if path is None:
            path = os.path.join(os.path.abspath(os.getcwd()), "")

        self.logger.info("Creating model definitions")

        src = os.path.join(os.path.split(__file__)[0], "") + "metadata_files"
        src_folders = os.listdir(src)
        for folder in src_folders:
            full_folder_name = os.path.join(src, folder)
            if os.path.isfile(full_folder_name) and not os.path.exists(
                os.path.join(path, folder)
            ):
                shutil.copy(full_folder_name, path)

        Path(f"{path}{MODEL_DEFINITIONS_PATH}/").mkdir(parents=True, exist_ok=True)

        self.logger.info(f"model directory initialized at {path}")

        logging.info("Creating feature engineering tasks directory")
        Path(f"{path}{FEATURE_ENGINEERING_PATH}/").mkdir(parents=True, exist_ok=True)

        self.logger.info(f"feature engineering tasks directory initialized at {path}")

    def read_repo_config(self):
        self.repo_config_rename()
        path = os.path.join(self.base_path, ".tmo/config.yaml")
        if os.path.exists(path):
            with open(path, "r") as handle:
                return yaml.safe_load(handle)

        self.logger.warning("ModelOps repo config doesn't exist")
        return None

    def write_repo_config(self, config, path=None):
        path = path if path else self.base_path
        config_dir = os.path.join(path, ".tmo")
        Path(config_dir).mkdir(parents=True, exist_ok=True)
        config_file = f"{config_dir}/config.yaml"

        with open(config_file, "w+") as f:
            yaml.dump(config, f, default_flow_style=False)

    def repo_config_exists(self, repo_path=None):
        path = repo_path if repo_path else self.base_path
        return Path(os.path.join(path, ".tmo")).is_file()

    def repo_config_rename(self, repo_path=None):
        path = repo_path if repo_path else self.base_path
        if Path(os.path.join(path, ".aoa")).is_file():
            os.rename(os.path.join(path, ".aoa"), os.path.join(path, ".tmo"))

    def train(self, model_id: str, data_conf: dict):
        trainer = TrainModel(self)
        trainer.train_model_local(
            model_id=model_id, rendered_dataset=data_conf, base_path=self.base_path
        )

    def evaluate(self, model_id: str, data_conf: dict):
        evaluator = EvaluateModel(self)
        evaluator.evaluate_model_local(
            model_id=model_id, rendered_dataset=data_conf, base_path=self.base_path
        )

    def batch_score_model_local(self, model_id: str, data_conf: dict):
        scorer = ScoreModel(self)
        scorer.batch_score_model_local(
            model_id=model_id, rendered_dataset=data_conf, base_path=self.base_path
        )

    @staticmethod
    def clone_repository(url, path, branch: str = "master"):
        repo = Repo.clone_from(url, path)
        repo.git.checkout(branch)

    @staticmethod
    def get_model_definitions_path() -> str:
        return MODEL_DEFINITIONS_PATH

    @staticmethod
    def get_feature_engineering_path() -> str:
        return FEATURE_ENGINEERING_PATH

    @staticmethod
    def _get_models_from_repo(templates_path: str = None) -> dict:
        if not os.path.isdir(templates_path):
            return {}

        models = {}

        for entity in sorted(os.listdir(templates_path)):
            if os.path.isfile(os.path.join(templates_path, entity)):
                continue

            model_metadata_dir = os.path.join(templates_path, entity)
            model_metadata_file = os.path.join(model_metadata_dir, MODEL_JSON)
            if os.path.isfile(model_metadata_file):
                with open(model_metadata_file, "r") as f:
                    model_metadata = json.load(f)
                    if model_metadata["language"] not in models:
                        models[model_metadata["language"]] = {}
                    models[model_metadata["language"]][model_metadata["id"]] = [
                        model_metadata["name"],
                        model_metadata_dir,
                    ]

        return models

    @staticmethod
    def _get_tasks_from_repo(templates_path: str = None) -> dict:
        if not os.path.isdir(templates_path):
            return {}

        tasks = {}

        for entity in sorted(os.listdir(templates_path)):
            from tmo.util.utils import get_files_in_path

            if os.path.isfile(os.path.join(templates_path, entity)):
                continue

            required_files = ["task.py", "requirements.txt"]
            task_dir = os.path.join(templates_path, entity)
            existing_files = get_files_in_path(task_dir)
            if set(required_files) == set(existing_files):
                tasks[entity] = task_dir

        return tasks



import os
from types import SimpleNamespace

import pytest

from tmo.util import artifacts


def test_save_plot_with_context_and_custom_dpi(tmp_path):
    out_dir = tmp_path / "out"
    out_dir.mkdir(parents=True, exist_ok=True)
    ctx = SimpleNamespace()
    ctx.artifact_output_path = str(out_dir)
    artifacts.save_plot("My Plot Title", dpi=300, context=ctx)  # type: ignore
    expected = os.path.join(ctx.artifact_output_path, "my_plot_title.png")
    assert os.path.isfile(expected)
    os.remove(expected)


def test_save_plot_without_context_uses_default_path():
    out_dir = os.path.join("artifacts", "output")
    os.makedirs(out_dir, exist_ok=True)
    artifacts.save_plot("Another Plot", dpi=150)
    expected = os.path.join(out_dir, "another_plot.png")
    assert os.path.isfile(expected)
    os.remove(expected)


def test_save_plot_propagates_exception_and_does_not_clear(monkeypatch, tmp_path):
    def raise_oserror(self, *args, **kwargs):  # noqa
        raise OSError("disk full")

    import matplotlib.figure as mf

    monkeypatch.setattr(mf.Figure, "savefig", raise_oserror)
    ctx = SimpleNamespace()
    ctx.artifact_output_path = str(tmp_path / "out")
    with pytest.raises(OSError):
        artifacts.save_plot("Bad Plot", dpi=72, context=ctx)  # type: ignore


def test_title_transformation_multiple_spaces():
    out_dir = os.path.join("artifacts", "output")
    os.makedirs(out_dir, exist_ok=True)
    artifacts.save_plot("  Mixed   CASE Title  ", dpi=10)
    expected_name = "__mixed___case_title__"
    expected = os.path.join(out_dir, f"{expected_name}.png")
    assert os.path.isfile(expected)
    os.remove(expected)


def test_save_plot_unicode_title(tmp_path):
    out_dir = tmp_path / "out"
    out_dir.mkdir(parents=True, exist_ok=True)
    ctx = SimpleNamespace()
    ctx.artifact_output_path = str(out_dir)
    title = "Título Ñá"
    artifacts.save_plot(title, dpi=72, context=ctx)  # type: ignore
    expected_name = title.replace(" ", "_").lower()
    expected = os.path.join(ctx.artifact_output_path, f"{expected_name}.png")
    assert os.path.isfile(expected)
    os.remove(expected)


def test_save_plot_with_subdir_in_title_creates_in_subdir():
    base = os.path.join("artifacts", "output")
    sub = os.path.join(base, "sub")
    os.makedirs(sub, exist_ok=True)
    title = "sub/plot name"
    artifacts.save_plot(title, dpi=50)
    expected_name = "plot_name"
    expected = os.path.join(base, "sub", f"{expected_name}.png")
    assert os.path.isfile(expected)
    os.remove(expected)


def test_save_plot_empty_title_raises(tmp_path):
    out_dir = tmp_path / "out"
    out_dir.mkdir(parents=True, exist_ok=True)
    ctx = SimpleNamespace()
    ctx.artifact_output_path = str(out_dir)
    try:
        artifacts.save_plot("", dpi=10, context=ctx)  # type: ignore
    except OSError:
        return
    files = [
        f
        for f in os.listdir(ctx.artifact_output_path)
        if os.path.isfile(os.path.join(ctx.artifact_output_path, f))
    ]
    assert files, "Expected a file to be created when title is empty"
    for f in files:
        os.remove(os.path.join(ctx.artifact_output_path, f))


def test_save_plot_passes_dpi_to_savefig(monkeypatch):
    import matplotlib.figure as mf

    captured = {}

    def fake_savefig(self, filename, dpi=None):
        captured["filename"] = filename
        captured["dpi"] = dpi

    monkeypatch.setattr(mf.Figure, "savefig", fake_savefig)
    artifacts.save_plot("My DPI Test", dpi=123)
    assert captured["dpi"] == 123
    assert captured["filename"].endswith("my_dpi_test")



import logging
import uuid
from typing import Optional

from tmo.types.base_dataset import BaseDatasetMixin
from tmo.types.dataset_metadata import (
    DataType,
    TypeEnum,
    CatalogType,
    CatalogBodyType,
    Variable,
    FeaturesEntityTargets,
    Predictions,
    Metadata,
    FeatureMetadata,
)

logger = logging.getLogger(__name__)

# Re-export classes from dataset_metadata for backward compatibility
__all__ = [
    "DataType",
    "TypeEnum",
    "CatalogType",
    "CatalogBodyType",
    "Variable",
    "FeaturesEntityTargets",
    "Predictions",
    "Metadata",
    "FeatureMetadata",
    "DatasetTemplate",
]


class DatasetTemplate(BaseDatasetMixin):
    id: uuid.UUID
    name: str
    description: str
    project_id: uuid.UUID
    owner_id: str
    catalog_type: CatalogType
    metadata: Metadata
    feature_metadata: FeatureMetadata
    entity: str
    target: list[str]

    def __init__(
        self,
        id: Optional[uuid.UUID] = None,
        name: Optional[str] = "SDK Dataset Template",
        description: Optional[str] = "VANTAGE dataset template",
        project_id: Optional[uuid.UUID] = None,
        owner_id: Optional[str] = None,
        catalog_type: Optional[CatalogType] = CatalogType.VANTAGE,
        metadata: Optional[Metadata] = Metadata(),
        feature_metadata: Optional[FeatureMetadata] = FeatureMetadata(),
    ):
        self.id = id
        self.name = name
        self.description = description
        self.project_id = project_id
        self.owner_id = owner_id
        self.catalog_type = catalog_type
        self.metadata = metadata
        self.feature_metadata = feature_metadata

    @property
    def id(self) -> uuid.UUID:
        return self._id

    @id.setter
    def id(self, value: uuid.UUID):
        if value is not None and not isinstance(value, uuid.UUID):
            try:
                value = uuid.UUID(value)  # noqa
            except ValueError:
                raise ValueError("ID must be a valid UUID.")
        self._id = value

    @property
    def project_id(self) -> uuid.UUID:
        return self._project_id

    @project_id.setter
    def project_id(self, value: uuid.UUID):
        if value is not None and not isinstance(value, uuid.UUID):
            try:
                value = uuid.UUID(value)  # noqa
            except ValueError:
                raise ValueError("Project ID must be a valid UUID.")
        self._project_id = value

    @property
    def owner_id(self) -> str:
        return self._owner_id

    @owner_id.setter
    def owner_id(self, value: str):
        if value is not None and not isinstance(value, str):
            raise ValueError("Owner ID must be a string.")
        self._owner_id = value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        if not isinstance(value, str):
            raise ValueError("Template name must be a string.")
        self._name = value

    @property
    def description(self) -> str:
        return self._description

    @description.setter
    def description(self, value: str):
        if not isinstance(value, str):
            raise ValueError("Description must be a string.")
        self._description = value

    @property
    def catalog_type(self) -> CatalogType:
        return self._catalog_type

    @catalog_type.setter
    def catalog_type(self, value: CatalogType):
        if not isinstance(value, CatalogType):
            raise ValueError("Catalog type must be an instance of CatalogType Enum.")
        self._catalog_type = value

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @metadata.setter
    def metadata(self, value: Metadata):
        if not self._isclass(value, Metadata):
            raise ValueError("Metadata must be an instance of Metadata class.")
        self._metadata = value

    @property
    def feature_metadata(self) -> FeatureMetadata:
        return self._feature_metadata

    @feature_metadata.setter
    def feature_metadata(self, value: FeatureMetadata):
        if not self._isclass(value, FeatureMetadata):
            raise ValueError(
                "FeatureMetadata must be an instance of FeatureMetadata class."
            )
        self._feature_metadata = value

    def get_df_template(self) -> dict:
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "catalogType": self.catalog_type,
            "entity": self.entity,
            "target": self.target,
            "features": self.features,
            "featuresQuery": self.features_query,
            "entityTargetsQuery": self.entity_targets_query,
            "predictionsQuery": self.predictions_query,
        }
